/*

  p_fptd.c      Compute first-passage-time density to threshold 1
                for the periodically modulated Ornstein-Uhlenbeck 
                process
                
  Algorithm     Solves renewal equation for the FPTD by block-by-
                block method, see Plesser, 1999, Appendix A

  CMEX version for direct use in MATLAB --- see p_fptd.m for calling syntax

  Copyright (C) 2000 Hans Ekkehard Plesser

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#define USAGE " Please type 'help p_fptd' for usage information! "

#include <math.h>
#include "mex.h"

/* The coefficients from eq. (A.8) are numerically difficult.
   Thus they have been tabulated using MAPLE.                 */
#include "../AlphaBetaGamma.h"      

/* Input Arguments */
#define	     MU_IN	prhs[0]
#define	      Q_IN	prhs[1]
#define	     OM_IN	prhs[2]
#define     PHI_IN      prhs[3]
#define	  SIGMA_IN	prhs[4]
#define	   TMAX_IN	prhs[5]
#define START_OPTS           6

/* Output Arguments */
#define	 RHO_OUT	plhs[0]
#define	   T_OUT	plhs[1]

/***********************************************************************/
/*                                                                     */
/*                  Macros                                             */
/*                                                                     */
/***********************************************************************/

#define MIN_NORM_DEF  0.9999   /* stop after this norm is reached      */
#define MIN_SEC_DEF   0.7      /* relative cut-off                     */
#define PTS_PER_PER   100      /* points per period                    */

#define	max(A, B)	((A) > (B) ? (A) : (B))
#define	min(A, B)	((A) < (B) ? (A) : (B))
#define odd(n)          ( (n) & 1 )

#define pi     3.14159265358979323846
#define root2  1.41421356237309504880

/* left hand side and kernel functions as macros  */
#define g(pre, e, ome2D, si, s0, omu, muvR)  \
  ( (pre) * exp( -pow( (omu) - (si) + (e)*((muvR)+(s0)), 2.0 ) / (ome2D) ) )

#define K(pre, e, omem, ome2D, si, sj)  \
  ( (pre) * exp( -pow( (omem) - (si) + (e) * (sj), 2.0 ) / (ome2D) ) )

/***********************************************************************/
/*                                                                     */
/*                  data structures                                    */
/*                                                                     */
/***********************************************************************/

typedef struct               /* input/problem parameters     */
{
  double mu,                 /* DC offset                    */
         q,                  /* AC amplitude                 */
         om, phi,            /* frequency and phase          */
         sigma,              /* noise amplitude              */
         v_R,                /* reset potential              */
         tmax,               /* time limit                   */
         min_norm,           /* norm criterion               */
         min_sec;            /* norm criterion per section   */
  long   n, nsec;            /* pts. total, per section      */
} par_type;

typedef struct               /* solution                     */
{
  double *t,                 /* time axis                    */
         *y;                 /* FPT distribution             */
} dat_type;

typedef struct               /* temporary quantities         */
{
  double h,                  /* step size                    */
         g_pref,             /* 7.5/sqrt(h)                  */
         D,                  /* sigma^2                      */
         omu,                /* 1 - mu                       */
         muvR,               /* mu - v_R                     */
         qfac,               /* q / sqrt(1+om^2)             */
         phase;              /* phi + acot om                */
  double *qs,                /* qfac sin(om*t[j]+phase)      */
         *e,                 /* exp(-t[j])                   */
         *omem,              /* (1-mu) (1-exp(-   t[j]))     */
         *ome2D,             /*      D (1-exp(- 2 t[j]))     */
         *pre_g,             /* g_pref /sqrt(1-exp(-2t[j]))  */
         *pre_K,             /* sqrt(t)/sqrt(1-exp(-2t[j]))  */
         eh,                 /* exp(-h/2)                    */
         omemh,              /* (1-mu) (1 - exp(-h/2))       */
         ome2Dh,             /*      D (1 - exp(-h)  )       */ 
         pre_Kh;             /* sqrt(h/2*(1-exp(-h)  )       */
} aux_type;

/***********************************************************************/
/*                                                                     */
/*                           Functions                                 */
/*                                                                     */
/***********************************************************************/

void init_calc(par_type *parp, dat_type *datp, aux_type *auxp,mxArray *plhs[]);
void do_calc(par_type *p, dat_type *d, aux_type *a);
void clear_calc(dat_type *d, aux_type *a);
void volterra(par_type *p, dat_type *d, aux_type *a,
	      long smin, long smax, double *nrm);


/***********************************************************************/
/*                                                                     */
/*                  interface function                                 */
/*                                                                     */
/***********************************************************************/

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])

{
  par_type pars, *parp = &pars;    /* allocate and get pointers to data */
  dat_type dats, *datp = &dats;    /* structures                        */
  aux_type auxs, *auxp = &auxs;
  char     opt_name[2];
  int      nopt;

  /* obtain data from input line */
  if ( nrhs < 6 || odd(nrhs)  || nlhs > 3 || nlhs < 1 )
    mexErrMsgTxt(USAGE);

  /* extract fixed parameters */
  parp->mu    = *mxGetPr(   MU_IN);
  parp->q     = *mxGetPr(    Q_IN);
  parp->om    = *mxGetPr(   OM_IN);
  parp->phi   = *mxGetPr(  PHI_IN);
  parp->sigma = *mxGetPr(SIGMA_IN);
  parp->tmax  = *mxGetPr( TMAX_IN);

  /* set defaults for option values */
  parp->v_R      = 0.0;           /* reset potential            */
  parp->n        = -1;            /* dummy value, see init calc */
  parp->nsec     = -1;            /* dummy value, see init calc */
  parp->min_norm = MIN_NORM_DEF;  /* norm criterion             */
  parp->min_sec  = MIN_SEC_DEF;   /* norm criterion per section */

  /* evaluate options */
  nopt = START_OPTS;
  while ( nopt < nrhs )
    {
      
      if (   mxGetString(prhs[nopt++], opt_name, 2) || nopt >= nrhs 
          || !mxIsDouble(prhs[nopt]) )
	mexErrMsgTxt(USAGE);

      switch ( opt_name[0] ) {
      case 'r':      /* reset potential    */
        parp->v_R      = mxGetScalar(prhs[nopt]);
	break;
      case 'N':      /* number of points   */
        parp->n        = mxGetScalar(prhs[nopt]);
	break;
      case 'n':      /* points per section */
        parp->nsec     = mxGetScalar(prhs[nopt]);
	break;
      case 'M':      /* minimal norm       */
        parp->min_norm = mxGetScalar(prhs[nopt]);
	break;
      case 'm':      /* min. norm per sec. */ 
        parp->min_sec  = mxGetScalar(prhs[nopt]);
	break;
      default:      /* wrong option       */
	fprintf(stderr, "Unrecognized option -%c\n", opt_name[0]);
	mexErrMsgTxt(USAGE);
	break;
      }   /* switch  */

      /* next option */
      nopt++;

    }  /* while */
     
  /* Do the actual computations in a subroutine */
  init_calc(parp, datp, auxp, plhs);

  do_calc(parp, datp, auxp);

  clear_calc(datp, auxp);

  return;
}


/***********************************************************************/
/*                                                                     */
/*                  functions                                          */
/*                                                                     */
/***********************************************************************/

void init_calc(par_type *parp, dat_type *datp, aux_type *auxp, 
               mxArray *plhs[])
{ 
  long   j;

  /* check for positive frequency */
  if ( parp->om <= 0.0 )
    mexErrMsgTxt("Frequency must be > 0.");

  /* compute number of points if not given on command line */
  if ( parp->n < 0 )
    parp->n = floor(0.5 * (double)PTS_PER_PER * parp->om * parp->tmax);

  /* check that number of points is odd and within size of AlphaBetaGamma */
  if ( !odd(parp->n) )
    parp->n++;
  if ( parp->n > N_COEFF )
    mexErrMsgTxt("Maximum number of points exceeded.");
 
  /* number of points per section if not given on command line */
  if ( parp->nsec < 0 )
    parp->nsec = parp->n - 1;  /* must be even */
  
  if ( odd(parp->nsec) )
    parp->nsec--;            

  /* make sure nsec does not exceed n */
  if ( parp->nsec >= parp->n )
    parp->nsec = parp->n - 1;

  /* convert frequency, phase to proper radians */
  parp->om  *= pi;
  parp->phi *= pi / 180;

  /* allocate memory for auxiliary arrays */
  auxp->qs    = (double *) calloc(parp->n, sizeof(double));
  auxp->e     = (double *) calloc(parp->n, sizeof(double));
  auxp->omem  = (double *) calloc(parp->n, sizeof(double));
  auxp->ome2D = (double *) calloc(parp->n, sizeof(double));
  auxp->pre_g = (double *) calloc(parp->n, sizeof(double));
  auxp->pre_K = (double *) calloc(parp->n, sizeof(double));

  /* Create a matrix or the return argument, get pointer */
  RHO_OUT = mxCreateDoubleMatrix(parp->n, 1, mxREAL);
  datp->y = mxGetPr(RHO_OUT);
  T_OUT   = mxCreateDoubleMatrix(parp->n, 1, mxREAL);
  datp->t = mxGetPr(  T_OUT);

  /* verify successful allocation */
  if (   !datp->t     || !datp->y     || !auxp->qs || !auxp->e 
      || !auxp->omem  || !auxp->ome2D || !auxp->pre_g 
      || !auxp->pre_K )
    mexErrMsgTxt("Error: Memory allocation failed.");

  /* compute auxiliary parameters */
  auxp->h      = parp->tmax / (double) (parp->n-1);   /* time step */
  auxp->g_pref = 7.5 / sqrt(auxp->h);
  auxp->D      = parp->sigma * parp->sigma;
  auxp->omu    = 1.0 - parp->mu;
  auxp->muvR   = parp->mu - parp->v_R;
  auxp->qfac   = parp->q / sqrt(1.0 + parp->om*parp->om);
  auxp->phase  = parp->phi + atan(1.0/parp->om);

  /* fill time array */
  for ( j = 0 ; j < parp->n ; j++ )
    datp->t[j]  = (double) j * auxp->h;

  datp->y[0] = 0.0;     /* initial condition */

  /* pre-computed quantities depending on t-s only, sin, exp, 
     weight arrays, i.e. for t==0 or t==s                     
     starting points only, rest computed on the fly          */
  auxp->qs[0]    = auxp->qfac * sin(auxp->phase);
  auxp->e[0]     = 1.0;
  auxp->omem[0]  = 0.0;
  auxp->ome2D[0] = 0.0;
  auxp->pre_g[0] = 0.0;           /* by definition  */
  auxp->pre_K[0] = 1.0 / root2;   /* limiting value */
  
  /* half-interval values */
  auxp->eh     = exp(-auxp->h/2.0);
  auxp->omemh  = - auxp->omu * expm1( -auxp->h/2.0 );
  auxp->ome2Dh = - auxp->D * expm1( -auxp->h );
  auxp->pre_Kh = pow(-0.5*auxp->h/expm1(-auxp->h), 0.5);

  return;

}  /* end -- init_calc */

/***********************************************************************/

void do_calc(par_type *p, dat_type *d, aux_type *a)
{
  double nrm;       /* cumulated sum of y                */
  long   smin,      /* last point before current section */
         j;        

  nrm  = 0.0;

  smin = 0;

  /* the big loop, calculating distribution section by section */
  /* p->nsec < p->n by def. so first section fits in any case  */
  do
    {
      /* compute auxiliary parameters, weight arrays for next section */
      for ( j = smin+1 ; j <= smin + p->nsec ; j++ )
	{
	  a->qs[j]    = a->qfac * sin(p->om*d->t[j] + a->phase);
	  a->e[j]     = exp( -d->t[j] );
	  a->omem[j]  = - a->omu * expm1( -d->t[j] );
	  a->ome2D[j] = - expm1( -2*d->t[j] );
	  a->pre_g[j] = pow(a->ome2D[j], -0.5);
	  a->pre_K[j] = pow(d->t[j], 0.5) * a->pre_g[j];
	  a->pre_g[j] *= a->g_pref;
	  a->ome2D[j] *= a->D;
	}  

      /* solve next section */
      volterra(p, d, a, smin, smin+p->nsec, &nrm);

      /* iterate */
      smin += p->nsec;
    }
  while (   smin + p->nsec < p->n
         && a->h*nrm       < p->min_norm
         && a->h*nrm       > p->min_sec  * (double) smin / (double) p->n );

  return;

}  /* end -- do_calc */

/***********************************************************************/

void clear_calc(dat_type *datp, aux_type *auxp)

{
  free(auxp->qs   );
  free(auxp->e    );
  free(auxp->omem );
  free(auxp->ome2D);
  free(auxp->pre_g);
  free(auxp->pre_K);
  return;

}  /* end -- clear_calc */

/***********************************************************************/

void volterra(par_type *p, dat_type *d, aux_type *a,
	      long smin, long smax, double *nrm)
{
  /* this function computes y(t) for t[smin+1]..t[smax] given
     y[0]..y[smin]                                            */
 
  long int  m, i;
  double    S1, S2, P, Q, R, T, V, W, G;

  /* outer loop -- points of desired function                               */
  /* NOTE: use m < smax-1 here to be on the save side even when smax is odd */
  /* NOTE: the m used here is 2m in Plesser, 1999                           */
  for ( m = smin ; m < smax-1 ; m += 2 )
    {
      /* the various kernel evaluations */
      P =            /* K(t[m+2], t[m]) */
        K(a->pre_K[2], a->e[2], a->omem[2], a->ome2D[2], a->qs[m+2], a->qs[m]);
      Q =            /* K(t[m+1], t[m+1/2]) */
        K(a->pre_Kh, a->eh, a->omemh, a->ome2Dh, a->qs[m+1], 
                        a->qfac * sin(p->om*(d->t[m]+0.5*a->h)+a->phase));
      R =            /* K(t[m+1], t[m]) */
        K(a->pre_K[1], a->e[1], a->omem[1], a->ome2D[1], a->qs[m+1], a->qs[m]);
      T =            /* K(t[m+2], t[m+1]) */
        K(a->pre_K[1], a->e[1], a->omem[1], a->ome2D[1],a->qs[m+2],a->qs[m+1]);
      V =            /* K(t[m+1], t[m+1]) */ 
        1.0/root2;  
      W =            /* K(t[m+2], t[m+2]) */ 
        1.0/root2;  

      /* determinant of iteration matrix */
      G = W*(Q+V)+Q*T/4.5;

      /* "fix" terms of right hand side */
      S1 = 
      g(a->pre_g[m+1],a->e[m+1],a->ome2D[m+1],a->qs[m+1],a->qs[0],
                                                         a->omu, a->muvR)
           - (R+3.0*Q) * d->y[m];
      S2 =
      g(a->pre_g[m+2],a->e[m+2],a->ome2D[m+2],a->qs[m+2],a->qs[0],
                                                         a->omu, a->muvR)
           - root2 * P * d->y[m];

      /* inner loop -- integral over known points */
      if ( m > 0 )
	{
	  S1 -=  Alpha[m+1] * d->y[0] * K(a->pre_K[m+1],a->e[m+1],a->omem[m+1],
                                          a->ome2D[m+1],a->qs[m+1],a->qs[0])
                +Beta[m]    * d->y[1] * K(a->pre_K[m  ],a->e[m  ],a->omem[m  ],
                                          a->ome2D[m  ],a->qs[m+1],a->qs[1])
                +Gamma[1]   * d->y[m] * R;

	  S2 -=  Alpha[m+2] * d->y[0] * K(a->pre_K[m+2],a->e[m+2],a->omem[m+2],
                                          a->ome2D[m+2],a->qs[m+2],a->qs[0])
                +Beta[m+1]  * d->y[1] * K(a->pre_K[m+1],a->e[m+1],a->omem[m+1],
                                          a->ome2D[m+1],a->qs[m+2],a->qs[1])
                +Gamma[2]   * d->y[m] * P;

	  for ( i = 2 ; i < m ; i += 2 )
	    {
	      S1 -= AlphaGamma[m+1-i] * d->y[i]   *             /* even */
                       K(a->pre_K[m+1-i],a->e[m+1-i],a->omem[m+1-i],
                         a->ome2D[m+1-i],a->qs[m+1],a->qs[i]);
	      S1 -= Beta[m-i]         * d->y[i+1] *             /* odd  */
                       K(a->pre_K[m  -i],a->e[m  -i],a->omem[m  -i],
                         a->ome2D[m  -i],a->qs[m+1],a->qs[i+1]);
	      
	      S2 -= AlphaGamma[m+2-i] * d->y[i]   *             /* even */
                       K(a->pre_K[m+2-i],a->e[m+2-i],a->omem[m+2-i],
                         a->ome2D[m+2-i],a->qs[m+2],a->qs[i]);
	      S2 -= Beta[m+1-i]       * d->y[i+1] *             /* odd  */
                       K(a->pre_K[m+1-i],a->e[m+1-i],a->omem[m+1-i],
                         a->ome2D[m+1-i],a->qs[m+2],a->qs[i+1]);
	    }
	}  /* end -- if m > 0 */

      /* compute the next two points */
      d->y[m+1] = ( W*S1/6.0   + Q    *S2/36.0/root2 ) / G;
      d->y[m+2] = (-T*S1/4.5   + (Q+V)*S2/ 6.0/root2 ) / G;

      /* stabilize by keeping non-negative, accumulate "norm" */
      if ( d->y[m+1] < 0.0 ) d->y[m+1] = 0.0;
      if ( d->y[m+2] < 0.0 ) d->y[m+2] = 0.0;

      *nrm += d->y[m+1]+d->y[m+2];

    }

  return;

}  /* end -- volterra */



      
      
