/*

  abel.c       Sample program for Block-by-Block algorithm for
               Abel eq. of 1st kind with square-root singularity

	       Literature: P. Linz
                           Analytical and Numerical Methods for
                             Volterra Equations
                           SIAM, Philadelphia, 1985
			   ch. 10.3

  The calling syntax is:

      abel t n outfile

  Computes the solution to a sample integral equation and writes it
  to outfile in two-column format

      int_0^t ds/sqrt(t-s) * K(t,s) * F(s) = g(t)

  Sample is 
      K(t, s) = 1 + t + s
      g(t)    = 2t/sqrt(t+1) * atanh(sqrt(t/t+1)) + 2*sqrt(t)
      F(0)    = 1

      exact solution Fe(t) = 1/1+t

  Input:       t  equation is solved on [0, t]
               n  number of points, must be odd
         outfile  name of output file

  Output: ASCII file       
            first column : time axis t
            second column: solution  F
            third column : exact solution Fe
           

  Documentation: Hans E. Plesser
                 Aspects of Signal Processing in Noisy Neurons
                 Ph.D. thesis, Universitaet Goettingen, 1999 
                 webdoc.sub.gwdg.de/diss/1999/plesser

  Hans E. Plesser, May 2000, plesser@chaos.gwdg.de

  Copyright (C) 2000 Hans Ekkehard Plesser

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>

/* The coefficients from Eq. (A.8) are numerically difficult.
   They have thus been tabulated using MAPLE.                 */
#include "AlphaBetaGamma.h"      


/***********************************************************************/
/*                                                                     */
/*                  Macros                                             */
/*                                                                     */
/***********************************************************************/

#define	max(A, B)	((A) > (B) ? (A) : (B))
#define	min(A, B)	((A) < (B) ? (A) : (B))
#define odd(n)          ( (n) & 1 )

#define pi     3.14159265358979323846
#define root2  1.41421356237309504880

/***********************************************************************/
/*                                                                     */
/*                  data structures                                    */
/*                                                                     */
/***********************************************************************/

typedef struct              /* input/problem parameters  */
{
  double tmax;              /* upper end of time range   */
  long   n;                 /* number of points          */
  char   outname[255];      /* file name                 */
} par_type;

typedef struct              /* solution                  */
{
  double *t,                /* time axis for convenience */
         *F;                /* function values           */
} dat_type;

typedef struct              /* auxiliary quantities      */
{
  double h,                 /* step size                 */
         g_pref;            /* prefactor of r.h.s.       */
} aux_type;

/***********************************************************************/
/*                                                                     */
/*                           Functions                                 */
/*                                                                     */
/***********************************************************************/

/* functions doing the work */
void init_calc(par_type *parp, dat_type *datp, aux_type *auxp);
void do_calc(par_type *p, dat_type *d, aux_type *a);
void save_calc(par_type *p, dat_type *d);
void clear_calc(dat_type *d, aux_type *a);

/* functions defining the right hand side, kernel and initial value    */
double g(double t);
double K(double t, double s);

/***********************************************************************/
/*                                                                     */
/*                  main program                                       */
/*                                                                     */
/***********************************************************************/

int main(int argc, char *argv[])
{
  par_type pars, *parp = &pars;
  dat_type dats, *datp = &dats;
  aux_type auxs, *auxp = &auxs;

  /* obtain data from input line */
  if ( argc != 4 )
    {
      printf("\n Usage: abel tmax num_points outfile.\n\n");
      exit(-1);
    }

  sscanf(argv[1], "%lf", &parp->tmax);
  sscanf(argv[2], "%ld", &parp->n);
  sprintf(parp->outname, "%s", argv[3]);

  /* make sure we have odd number of points */
  if ( !odd(parp->n) )
    {
      printf("\n Error: number of points must be odd.\n\n");
      exit(-1);
    }

  /* make sure that we do not exceed number of coefficients included
     from AlphaBetaGamma.h                                            */
  if ( parp->n > N_COEFF )
    {
      printf("\n Error: coefficients tabulated for <= %d points only.\n\n",
             N_COEFF);
      exit(-1);
    }

  init_calc(parp, datp, auxp);    /* initialize data structures */
  do_calc(parp, datp, auxp);      /* calculate solution         */
  save_calc(parp, datp);          /* write to file              */
  clear_calc(datp, auxp);         /* clean up                   */

  return 0;
}


/***********************************************************************/
/*                                                                     */
/*                  functions                                          */
/*                                                                     */
/***********************************************************************/

void init_calc(par_type *parp, dat_type *datp, aux_type *auxp)

{ 
  long j;

  /* allocate memory for data array */
  datp->t     = (double *) calloc(parp->n, sizeof(double));
  datp->F     = (double *) calloc(parp->n, sizeof(double));

  /* compute auxiliary parameters */
  auxp->h      = parp->tmax / (double) (parp->n-1);   /* time step    */
  auxp->g_pref = 7.5 / sqrt(auxp->h);                 /* rhs factor   */

  /* fill time array */
  for ( j = 0 ; j < parp->n ; j++ )
    datp->t[j]  = (double) j * auxp->h;

  datp->F[0] = 1.0;                          /* set initial condition */

  return;
}

/*********************************************************************/

void do_calc(par_type *p, dat_type *d, aux_type *a)

     /* This routine performs the actual computations. */
        

{

  long int  m, i;
  double    S1, S2, th, P, Q, R, T, V, W, G;

  /* outer loop -- points of desired function */
  /* NOTE: the variable m here is equivalent to 2m in Appendix A
           of Plesser, 1999                                       */
  for ( m = 0 ; m < p->n-1 ; m += 2 )
    {
      /* interpolated time point */
      th = d->t[m] + 0.5*a->h;

      /* various kernel evaluations */
      P = K(d->t[m+2], d->t[m]);
      Q = K(d->t[m+1], th);
      R = K(d->t[m+1], d->t[m]);
      T = K(d->t[m+2], d->t[m+1]);
      V = K(d->t[m+1], d->t[m+1]);
      W = K(d->t[m+2], d->t[m+2]);

      /* determinant of iteration matrix */
      G = W*(Q+V) + Q*T/4.5;

      /* "fixed" terms of right hand side */
      S1 = a->g_pref * g(d->t[m+1]) - (R+3.0 * Q) * d->F[m];
      S2 = a->g_pref * g(d->t[m+2]) -  root2 * P  * d->F[m];

      /* integral over known points */
      if ( m > 0 )
	{
	  /* Split of a few points at beginning/end of integral which
             have slightly different weights. Not really necessary,
             but makes rest simpler.                                 */
	  S1 -=   Alpha[m+1] * K(d->t[m+1], d->t[0]) * d->F[0]
                + Beta[m]    * K(d->t[m+1], d->t[1]) * d->F[1]
                + Gamma[1]   * R                     * d->F[m];
	  S2 -=   Alpha[m+2] * K(d->t[m+2], d->t[0]) * d->F[0]
                + Beta[m+1]  * K(d->t[m+2], d->t[1]) * d->F[1]
                + Gamma[2]   * P                     * d->F[m];

	  for ( i = 2 ; i < m ; i += 2 )
	    {
	      S1 -= AlphaGamma[m+1-i] * K(d->t[m+1], d->t[i])   * d->F[i]; 
	      S1 -= Beta[m-i]         * K(d->t[m+1], d->t[i+1]) * d->F[i+1];
	      
	      S2 -= AlphaGamma[m+2-i] * K(d->t[m+2], d->t[i])   * d->F[i];
	      S2 -= Beta[m+1-i]       * K(d->t[m+2], d->t[i+1]) * d->F[i+1];
	    }  /* end -- for i */
	}  /* end -- if m > 0 */

      /* compute the next two points */
      d->F[m+1] = ( W*S1/6.0   + Q    *S2/36.0/root2 ) / G;
      d->F[m+2] = (-T*S1/4.5   + (Q+V)*S2/ 6.0/root2 ) / G;

    }  /* end -- for m */

  return;

}  /* end -- do_calc */

/***********************************************************************/

void save_calc(par_type *p, dat_type *d)

     /* write result into ASCII file */

{
  FILE    *outfile;
  long     j;

  /* open output file */
  outfile = fopen(p->outname, "w");
  if ( !outfile )
    {
      printf("\n Error: Could not open %s!\n\n", p->outname);
      exit(-11);
    }

  /* write data line by line */
  for ( j = 0 ; j < p->n ; j++ )
    fprintf(outfile, "%20.15e %20.15e %20.15e\n", 
            d->t[j], d->F[j], 1.0 / (1.0 + d->t[j]));

  fclose(outfile);

  return;

}  /* end -- save_calc */

/***********************************************************************/

void clear_calc(dat_type *datp, aux_type *auxp)

     /* free memory */

{
  free(datp->t);
  free(datp->F);
  return;
}



/**********************************************************************

  Below are rhs, kernel and initial value function
  Replace as needed

 **********************************************************************/

double g(double t)

     /* must be continuous on 0..tmax */

{
  return  2.0*t/sqrt(t+1.0) 
         *0.5*log((sqrt(t+1.0)+sqrt(t))/(sqrt(t+1.0)-sqrt(t)))
         +2.0*sqrt(t);
}

double K(double t, double s)

     /* must be continuous for 0 <= s <= t <= tmax;
        if necessary, use finite limit for s==t      */

{
  return 1.0+t+s;
}


