/*

  ap_fptd.c     Compute first-passage-time density to threshold 1
                for the aperiodically modulated Ornstein-Uhlenbeck 
                process
                
  Algorithm     Solves renewal equation for the FPTD by block-by-
                block method, see Plesser, 1999, Appendix A

  Copyright (C) 2000 Hans Ekkehard Plesser

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


  The calling syntax is:

*/

#define USAGE \
"                                                                      \n"\
" First-passage-time density for periodic stimuli                      \n"\
"                                                                      \n"\
"  ap_fptd [options] mu q sigma stimfile outfile                       \n"\
"                                                                      \n"\
"  Input:         mu  DC offset                                        \n"\
"                  q  rms amplitude of AC input (Plesser&Gerstner 2000)\n"\
"              sigma  rms amplitude of noise			       \n"\
"           stimfile  MATLAB stimulus file from ap_stim.m              \n"\
"            outfile  MATLAB file containing rho(t), params	       \n"\
"								       \n"\
"  Options:   -r ###  reset potential   [ 0 ]	    	               \n"\
"                                                                      \n"\
"             -M ###  terminate if norm > ### [ 0.9999 ]               \n"\
"             -Z      gzip output file [ do not gzip ]                 \n"\
"                                                                      \n"\
"  Note:      Negative functions values arising from numerical noise   \n"\
"             are set to zero to stabilize the computation.            \n"\
"                                                                      \n"

/*
  File formats: 

  Stimulus:  MATLAB file containing variables 

               dt     scalar
               F_n    vector 
               var_n  scalar (not required by ap_fptd)
               I_n    vector (not required by ap_fptd)

	     see ap_stim.m for details

  Solution:  rho(j) = rho(2*(j-1)*dt) since intermediate stimulus points
             are needed for midpoint (nh+1/2) in block-by-block method.

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <string.h>
#include <getopt.h>

#include <mat.h>     /* MATLAB file handling */

/* The coefficients from eq. (A.8) are numerically difficult.
   Thus they have been tabulated using MAPLE.                 */
#include "AlphaBetaGamma.h"      

/***********************************************************************/
/*                                                                     */
/*                  Macros                                             */
/*                                                                     */
/***********************************************************************/

#define	max(A, B)	((A) > (B) ? (A) : (B))
#define	min(A, B)	((A) < (B) ? (A) : (B))
#define odd(n)          ( (n) & 1 )

#define pi     3.14159265358979323846
#define root2  1.41421356237309504880

#define MIN_NORM_DEF 0.99999

#define FILE_NAME_LENGTH 255

/***********************************************************************/
/*                                                                     */
/*                  data structures                                    */
/*                                                                     */
/***********************************************************************/

typedef struct              /* input/problem parameters     */
{
  long   n,                 /* number of points in solution */
         n_in;              /* number of points in signal   */
  double dt_in,             /* time step of input signal    */
         *iS,               /* integrated input signal      */
         mu,                /* DC offset                    */
         q,                 /* rms amp / sqrt(2)            */
         sigma,             /* noise amplitude              */
         D,                 /* noise intensity              */
         v_R,               /* reset potential              */
         min_norm;          /* norm criterion               */
  int    zip;               /* compress?                    */
  char   *stimname,         /* stimulus file name           */
         *outname;          /* name of output file          */
} par_type;

typedef struct              /* solution                     */
{
  double *t,                /* time axis of output          */
         *y;                /* function values              */
} dat_type;

typedef struct              /* auxiliary quantities         */
{
  double h,                 /* step size                    */
         g_pref;            /* prefactor of r.h.s.          */
} aux_type;

/***********************************************************************/
/*                                                                     */
/*                           Functions                                 */
/*                                                                     */
/***********************************************************************/

void init_calc(par_type *parp, dat_type *datp, aux_type *auxp);
void do_calc(par_type *p, dat_type *d, aux_type *a);
void save_calc(par_type *p, dat_type *d);
void clear_calc(par_type *p, dat_type *d, aux_type *a);

/* functions defining the right hand side and kernel of the equation */
/* call with array indices, to have straight access to integrate signal */
double g(long t_j, par_type *parp);
double K(long t_j, long s_k, par_type *parp);

/* simplify MATLAB matrix handling */
long myMatGet(MATFile *fp, const char *name, double **data);
void myMatPut(MATFile *fp, char *name, long rows, long cols, double *data);


/***********************************************************************/
/*                                                                     */
/*                  main program                                       */
/*                                                                     */
/***********************************************************************/

int main(int argc, char *argv[])
{
  par_type pars, *parp = &pars;
  dat_type dats, *datp = &dats;
  aux_type auxs, *auxp = &auxs;
  int      errflg;
  char     c; 

  /* set defaults for options */
  parp->v_R      = 0.0;           /* reset potential            */
  parp->min_norm = MIN_NORM_DEF;  /* norm criterion             */
  parp->zip      = 0;             /* do not zip                 */

  /* process options */
  errflg = 0;
  opterr = 0;   /* errors handled by program */
  
  while (   !errflg 
         && ( c=getopt(argc,argv,":r:M:m:Zh") ) != -1 )
    switch ( c ) {
    case 'r':      /* reset potential    */
      sscanf(optarg, "%lf", &parp->v_R);
      break;
    case 'M':      /* minimal norm       */
      sscanf(optarg, "%lf", &parp->min_norm);
      break;
    case 'Z':      /* gzip output        */
      parp->zip = 1;
      break;
    case 'h':      /* help               */
      errflg++;
      break;
    case ':':      /* missing option     */
      fprintf(stderr, "Option -%c requires an argument\n", optopt);
      errflg++;
      break;
    case '?':      /* wrong option       */
      fprintf(stderr, "Unrecognized option -%c\n", optopt);
      errflg++;
      break;
    default:
      fprintf(stderr, "This error is impossible\n");
      errflg++;
      break;
    }   /* switch, while */

  /* now read the required input arguments if okay */
  if ( optind + 5 > argc || errflg )
    {
      fprintf(stderr, USAGE);
      exit(-1);
    }

  sscanf(argv[optind++], "%lf", &parp->mu);
  sscanf(argv[optind++], "%lf", &parp->q);
  sscanf(argv[optind++], "%lf", &parp->sigma);
  parp->stimname = argv[optind++];
  parp->outname  = argv[optind++];

  init_calc(parp, datp, auxp);
  do_calc(parp, datp, auxp);
  save_calc(parp, datp);
  clear_calc(parp, datp, auxp);

  return 0;
}


/***********************************************************************/
/*                                                                     */
/*                  functions                                          */
/*                                                                     */
/***********************************************************************/

void init_calc(par_type *parp, dat_type *datp, aux_type *auxp)
{ 
  long     j;
  double   *tmp;
  char     inname[FILE_NAME_LENGTH];
  MATFile  *infile;


  /* open stimulus file */
  sprintf(inname, "%s.mat", parp->stimname);
  infile = matOpen(inname, "r");
  if ( !infile )
    {
      fprintf(stderr, "\n Error: Could not matOpen %s!\n\n", parp->stimname);
      exit(-11);
    }
  
  /* get input time step */
  myMatGet(infile, "dt", &tmp);

  /* get input */
  parp->n_in = myMatGet(infile, "F_n", &(parp->iS));
  
  if ( !tmp || !parp->iS )
    {
      fprintf(stderr, "\n Error: Could not read from %s!\n\n", parp->stimname);
      exit(-12);
    }
  else
    {
      parp->dt_in = *tmp;
      free(tmp);
    }

  matClose(infile);

  /* get number of output points, time axis --- 
       one output point for every two input points */
  if ( odd(parp->n_in) )
    parp->n = (parp->n_in-1)/2 - 1;
  else
    parp->n = parp->n_in/2 - 1;

  /* check that we have enough coefficients */
  if ( parp->n > N_COEFF )
    {
      fprintf(stderr, "\n\n Error: maximum %d points!\n\n", N_COEFF);
      exit(-1);
    }

  /* allocate memory for data and auxiliary arrays */
  datp->t     = (double *) calloc(parp->n, sizeof(double));
  datp->y     = (double *) calloc(parp->n, sizeof(double));

  /* verify successful allocation */
  if (   !datp->t     || !datp->y     )
    {
      fprintf(stderr,
	      "\n\n Error: Memory allocation failed (%d MB required).\n\n",
              (int) ceil(2.0*parp->n*sizeof(double)/1048576.0));
      exit(-1);
    }

  /* compute auxiliary parameters */
  auxp->h      = 2.0 * parp->dt_in;           /* time step */
  auxp->g_pref = 7.5 / sqrt(auxp->h);
  parp->D      = parp->sigma * parp->sigma;

  /* fill time array */
  for ( j = 0 ; j < parp->n ; j++ )
    datp->t[j]  = (double) j * auxp->h;

  datp->y[0] = 0;   /* initial condition */

  return;
}

/***********************************************************************/

void do_calc(par_type *p, dat_type *d, aux_type *a)

     /* This routine performs the actual computations. */

     /* NOTE: arrays in d have normal spacing, but indices
              to signal-related arrays, funcs have double spacing */

{

  long int  m, i;
  double    S1, S2, P, Q, R, T, V, W, G, nrm;

  nrm = 0.0;

  /* outer loop -- points of desired function */
  for ( m = 0 ; m < p->n-2 ; m += 2 )
    {
      /* various kernel evaluations */
      P = K(2*(m+2), 2*m, p);
      Q = K(2*(m+1), 2*m + 1, p);   /* interpolation at half-interval */
      R = K(2*(m+1), 2*m, p);
      T = K(2*(m+2), 2*(m+1), p);
      V = K(2*(m+1), 2*(m+1), p);
      W = K(2*(m+2), 2*(m+2), p);

      /* determinant of iteration matrix */
      G = W*(Q+V)+Q*T/4.5;

      /* "fix" terms of right hand side */
      S1 = a->g_pref * g(2*(m+1), p) - (R+3.0*Q)  * d->y[m];
      S2 = a->g_pref * g(2*(m+2), p) - root2 *  P * d->y[m];

      /* integral over known points */
      if ( m > 0 )
	{
	  /* Split of a few points at beginning/end of integral which
             have slightly different weights. Not really necessary
             for the beta terms, but makes rest simpler.             */
	  S1 -=   Alpha[m+1] * K(2*(m+1), 0, p) * d->y[0]
                + Beta[m]    * K(2*(m+1), 2, p) * d->y[1]
                + Gamma[1]   * R                * d->y[m];
	  S2 -=   Alpha[m+2] * K(2*(m+2), 0, p) * d->y[0]
                + Beta[m+1]  * K(2*(m+2), 2, p) * d->y[1]
                + Gamma[2]   * P                * d->y[m];

	  for ( i = 2 ; i < m ; i += 2 )
	    {
	      S1 -= AlphaGamma[m+1-i]                      /* even */
		      * K(2*(m+1), 2*i, p)   * d->y[i]; 
	      S1 -= Beta[m-i]                              /* odd  */
		      * K(2*(m+1), 2*(i+1), p) * d->y[i+1];
	      
	      S2 -= AlphaGamma[m+2-i]                      /* even */
		      * K(2*(m+2), 2*i, p)   * d->y[i];
	      S2 -= Beta[m+1-i]                            /* odd  */
		      * K(2*(m+2), 2*(i+1), p) * d->y[i+1];
	    }  /* end -- for i */
	}  /* end -- if m > 0 */

      /* compute the next two points */
      d->y[m+1] = ( W*S1/6.0   + Q    *S2/36.0/root2 ) / G;
      d->y[m+2] = (-T*S1/4.5   + (Q+V)*S2/ 6.0/root2 ) / G;

      /* stabilize by keeping non-negative, compute norm */
      if ( d->y[m+1] < 0.0 ) d->y[m+1] = 0.0;
      if ( d->y[m+2] < 0.0 ) d->y[m+2] = 0.0;
      nrm += d->y[m+1]+d->y[m+2];
      
      if ( nrm * a->h > p->min_norm )
	break;

    }  /* end -- for m */

  /* fill remaining y entries with zeros, in case there are
     leftovers from earlier runs                              */
  for ( i = m+3 ; i < p->n ; i++ ) 
    d->y[i] = 0.0;

  return;

}  /* end -- do_calc */

/***********************************************************************/

void save_calc(par_type *p, dat_type *d)
{
  MATFile  *outfile;
  char     outname[FILE_NAME_LENGTH],
           zip_cmd[FILE_NAME_LENGTH];

  sprintf(outname, "%s.mat", p->outname);
  outfile = matOpen(outname, "w");
  if ( !outfile )
    {
      fprintf(stderr, "\nCould not matOpen %s !\n", p->outname);
      exit(1);
    }

  myMatPut(outfile, "t",   p->n, 1, d->t);   
  myMatPut(outfile, "rho", p->n, 1, d->y);
  myMatPut(outfile, "mu",      1, 1, &p->mu);
  myMatPut(outfile, "q",       1, 1, &p->q);
  myMatPut(outfile, "sigma",   1, 1, &p->sigma);
  myMatPut(outfile, "v_R",     1, 1, &p->v_R);

  matClose(outfile);

  /* compress */
  if ( p->zip )
    {
      sprintf(zip_cmd, "gzip %s.mat", p->outname);
      system(zip_cmd);
    }

  return;
}

/***********************************************************************/

void clear_calc(par_type *parp, dat_type *datp, aux_type *auxp)

{
  free(parp->iS   );
  free(datp->t    );
  free(datp->y    );
  return;
}

/***********************************************************************/

/*  Now follow MATLAB matrix handling functions                        */

/***********************************************************************/

long myMatGet(MATFile *fp, const char *name, double **data)

/* places r x c matrix under name 'name' in file fp, takes data from data */

{
  long    rows, cols;
  mxArray *mdata;

  /* read MATLAB data structure */
  mdata = matGetArray(fp, name);
  if ( !mdata )
    {
      *data = NULL;   /* read error */
      return 0;
    }

  rows = mxGetM(mdata);
  cols = mxGetN(mdata);
  if ( rows > 1 && cols > 1 )
    {
      *data = NULL;
      return 0;
    }

  if ( cols > 1 )
    rows = cols;

  /* allocate memory, works 1D only!!! */
  *data = malloc(rows*sizeof(double));
  if ( !*data )
    return 0;

  /* copy data */
  memcpy(*data, mxGetPr(mdata), rows*sizeof(double));

  mxDestroyArray(mdata);

  return rows;
}  /* end -- myMatGet */


/**********************************************************************/

void myMatPut(MATFile *fp, char *name, long rows, long cols, double *data)

/* places r x c matrix under name 'name' in file fp, takes data from data */

{
  mxArray *mdata;
 
  mdata = mxCreateDoubleMatrix(rows, cols, mxREAL);
  memcpy(mxGetPr(mdata), data, rows*cols*sizeof(double));
  mxSetName(mdata, name);
  matPutArray(fp, mdata);
  mxDestroyArray(mdata);

  return;
}  /* end -- myMatPut */


/**********************************************************************

  Below are rhs, kernel and initial value function

  For computation of v from Sint == iS == F, see
    Plesser and Gerstner, Neural Comput 12:367-384 (2000), Sec. 2.3 

 **********************************************************************/

double g(long j, par_type *p)

     /* must be continuous on 0..tmax */

{
  double v, sig2, f, t;

  if ( j > 0 )
    {
      t    = (double) j * p->dt_in;
      v    =   p->v_R   * exp(-t)
             - p->mu  * expm1(-t)
	     + p->q     * (p->iS[j] - exp(-t) * p->iS[0]);
      sig2 = - p->D * expm1(-2.0*t);
      f    = exp(-pow(1.0-v, 2.0)/sig2) / sqrt(pi*sig2); 
    }
  else
    f = 0.0;

  return f;
}

double K(long j, long k, par_type *p)

     /* must be continuous for 0 <= s <= t <= tmax,
        if required used finite limit for s==t      */

{
  double v, sig2, dt, f;

  if ( j != k )
    {
      dt   = (double) (j-k) * p->dt_in;
      v    = exp(-dt) - p->mu  * expm1(-dt)
	     + p->q * (p->iS[j] - exp(-dt) * p->iS[k]);
      sig2 = - p->D * expm1(-2.0*dt);
      f    = sqrt(dt) * exp(-pow(1.0-v, 2.0)/sig2) / sqrt(pi*sig2);  
    } 
  else
    f = 1.0 / sqrt(2.0 * pi * p->D);   /* limiting value */

  return f;

}





      
      
