function [I_n, F_n, var_n, t] = ap_stim(log2_n, dt, om_c, IF_file),

%   [I_n, F_n, var_n, t] = ap_stim(log2_n, dt, om_c [, IF_file])
%
%   Create aperiodic stimulus as defined in
%   H. E. Plesser, Dissertation, Goettingen, 1999, Ch. 2.3
%
%   Input:  
%      log2_n    stimulus length n = 2^log2_n
%      dt        stimulus time step
%      om_c      cut-off frequency for amplitudes, see Eq. (17)
%      IF_file   output file for stimulus, for use with ap_fptd, ap_train
%                   no file is written if no name is given
%
%   Output:
%      I_n       normalized stimulus
%      F_n       normalized stimulus "integral"
%      var_n     normalized stimulus variance
%      t         stimulus time axis
%
%   Stimulus as in Eq. (2.39):  
%          I(t)      = mu                 + q   * I_n
%
%   Membrane potential as in Eq. (2.40): 
%        v_0(t|0, 0) = mu * ( 1-exp(-t) ) + q   * ( F_n - exp(-t)*F_n(1) )
%
%   Membrane potential variance as above Eq. (2.41):
%   <(delta v_0)^2>  =                      q^2 * var_n
%
%   IF_file is a MATLAB file containing I_n, F_n, var_n, dt
%
%   NOTE: amplitude q differs by sqrt(2) from Plesser&Gerstner (2000)

%   Copyright (C) 2000 Hans Ekkehard Plesser
%
%   This program is free software; you can redistribute it and/or
%   modify it under the terms of the GNU General Public License
%   as published by the Free Software Foundation; either version 2
%   of the License, or (at your option) any later version.
%
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with this program; if not, write to the Free Software
%   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

%  The sums in Eq. (2.39) and the equation below Eq. (2.40) are evaluated
%  using MATLAB's built-in complex FFT function.  To obtain the required
%  cosine transformation, we use (see also Numerical Recipes):
%
%  sum_{j=1}^n alpha_j cos(j dom t + phi_j) 
%        
%         = 0.5 * sum_{j=-n}^n d_j exp(-I psi_j) * exp(-I j dom t - I phi_j)
%  
%  with  j < 0 : d_j = alpha_{-j}, psi_j = - phi_j
%        j = 0 : d_j = 0         , psi_j =   0      (arbitrary)
%        j > 0 : d_j = alpha_j   , psi_j =   phi_j
%
%  The stimulus thus contains half as many modes as data points. Note 
%  that due to MATLABs (standard) way of handling FFT arguments, 
%  j=0..n fill the first part of the array, followed by j=-(n-1)..-1,
%  while j=-n has to be added explicitly.

I    = sqrt(-1);
n    = 2^floor(log2_n);       % length of stimulus
N    = n/2;                   % number of modes in stimulus
dom  = 2*pi / (dt * n);       % frequency interval between modes
ncut = ceil(om_c / dom);      % cut-off bin for amplitudes
if ncut > N,
  ncut = N;
end

% create normalized amplitudes, Eq. (2.41), Eq. (2.39)
alpha   = [              0 ;  ...             % mean-free stimulus
           ones(ncut-1, 1) ;  ...             % plateau
	   exp(-0.5*((ncut:N).'-ncut).^2)];   % soft cut-off
alpha_n = alpha / sqrt(sum(alpha.^2));

% create random phases
phi   = [ 0 ; 2*pi*rand(N, 1) ];


% compute normalized stimulus, cf. Eq. (2.39)

% combine---sign in the exponent is important!
r_n = alpha_n .* exp(-I * phi);       

% complex FFT 
% NOTE: r_n is N+1 long; add j = -n explicitly
% take real to eliminate numerical noise in imaginary part
I_n = 0.5 * real(  fft( [r_n ; conj(r_n(N:-1:2))] ) ...
                 + (-1).^(0:(n-1)).' * conj(r_n(N+1)) );


% compute "integrated" stimulus, cf. Eq. (2.40)

% dampened amplitudes
alpha_d = alpha_n ./ sqrt( 1 + ((0:N).' * dom).^2);

% shifted phases, extra pi/2 to shift from cos to sin
phi_d   = phi + [ 0 ; acot( (1:N).' * dom )-pi/2 ];
  
% combine---sign in the exponent is important!
r_d = alpha_d .* exp(-I * phi_d);

% complex FFT
% NOTE: r_d is N+1 long; add j = -n explicitly
% take real to eliminate numerical noise in imaginary part
F_n = 0.5 * real(  fft( [r_d ; conj(r_d(N:-1:2))] ) ...
                 + (-1).^(0:(n-1)).' * conj(r_d(N+1)) );


% normalized variance, cf. above Eq. (2.41)
var_n = 0.5 * sum(alpha_d.^2);


% time axis, if requested
if nargout > 3,
  t = (0:(n-1)).' * dt;
end


% save stimulus to file, if requested
if nargin > 3,
  save(IF_file, 'I_n', 'F_n', 'var_n', 'dt');
end

return
