/*

  p_fptd.c      Compute first-passage-time density to threshold 1
                for the periodically modulated Ornstein-Uhlenbeck 
                process
                
  Algorithm     Solves renewal equation for the FPTD by block-by-
                block method, see Plesser, 1999, Appendix A

  Copyright (C) 2000 Hans Ekkehard Plesser

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


  The calling syntax is:

*/

#define USAGE \
"                                                                      \n"\
" First-passage-time density for periodic stimuli                      \n"\
"                                                                      \n"\
"  p_fptd [options] mu q Om phi sigma tmax outfile                     \n"\
"                                                                      \n"\
"  Input:         mu  DC input                                         \n"\
"                  q  amplitude of AC input			       \n"\
"                 Om  stimulus frequency / pi [ in radians ]           \n"\
"                phi  stimulus phase at t = 0 [ in degrees ]           \n"\
"              sigma  rms amplitude of noise			       \n"\
"               tmax  time limit for computation		       \n"\
"            outfile  MATLAB file containing rho(t), parameters	       \n"\
"								       \n"\
"  Options:   -r ###  reset potential  [ default: 0 ]	               \n"\
"                                                                      \n"\
"             -N ###  number of points [ 100 per period ]              \n"\
"             -n ###  number of points per section [ number of points ]\n"\
"             -M ###  terminate if norm > ### [ 0.9999 ]               \n"\
"             -m ###  terminate if norm < ###*curr_pts/tot_pts [ 0.7 ] \n"\
"             -Z      gzip output file [ do not gzip ]                 \n"\
"                                                                      \n"\
"             When the option -n ### is used, the calculation is split \n"\
"             into sections of ### points.  The calculation is termi-  \n"\
"             nated after a section if either the minimum total norm   \n"\
"             has been reached (default 0.9999, set with -M ###), or   \n"\
"             the norm is less than 0.7 (change with -m ###) of the    \n"\
"             fraction of points computed so far (i.e. one is to expect\n"\
"             that if the compuation were continued over the total     \n"\
"             number of points, only a norm of less than 0.7 would be  \n"\
"             obtained.  Since CPU time grows quadratically with number\n"\
"             of computed points, sectioning is STRONGLY advised!      \n"\
"                                                                      \n"\
"  Note:      Negative functions values arising from numerical noise   \n"\
"             are set to zero to stabilize the computation.            \n"\
"                                                                      \n"


#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include <string.h>
#include <getopt.h>

#include <mat.h>       /* MATLAB file handling */

/* The coefficients from eq. (A.8) are numerically difficult.
   Thus they have been tabulated using MAPLE worksheet bb_tabulate.mws */
#include "AlphaBetaGamma.h"      


/***********************************************************************/
/*                                                                     */
/*                  Macros                                             */
/*                                                                     */
/***********************************************************************/

#define MIN_NORM_DEF  0.9999   /* stop after this norm is reached      */
#define MIN_SEC_DEF   0.7      /* relative cut-off                     */
#define PTS_PER_PER   100      /* point per period                     */

#define FILE_NAME_LENGTH 255   /* max length of file names             */

#define	max(A, B)	((A) > (B) ? (A) : (B))
#define	min(A, B)	((A) < (B) ? (A) : (B))
#define odd(n)          ( (n) & 1 )

#define pi     3.14159265358979323846
#define root2  1.41421356237309504880

/* left hand side and kernel functions as macros      */
/* for arguments, see typedef struct { ... } aux_type */
#define g(pre, e, ome2D, si, s0, omu, muvR)  \
  ( (pre) * exp( -pow( (omu) - (si) + (e)*((muvR)+(s0)), 2.0 ) / (ome2D) ) )

#define K(pre, e, omem, ome2D, si, sj)  \
  ( (pre) * exp( -pow( (omem) - (si) + (e) * (sj), 2.0 ) / (ome2D) ) )

/***********************************************************************/
/*                                                                     */
/*                  data structures                                    */
/*                                                                     */
/***********************************************************************/

typedef struct               /* input/problem parameters     */
{
  double mu,                 /* DC offset                    */
         q,                  /* AC amplitude                 */
         om, phi,            /* frequency and phase          */
         sigma,              /* noise amplitude              */
         v_R,                /* reset potential              */
         tmax,               /* time limit                   */
         min_norm,           /* norm criterion               */
         min_sec;            /* norm criterion per section   */
  long   n, nsec;            /* pts. total, per section      */
  int    zip;                /* shall we g'zip?              */
  char   *outname;           /* output file name             */
} par_type;

typedef struct               /* solution                     */
{
  double *t,                 /* time axis                    */
         *y;                 /* FPT distribution             */
} dat_type;

typedef struct               /* temporary quantities         */
{
  double h,                  /* step size                    */
         g_pref,             /* 7.5/sqrt(h)                  */
         D,                  /* sigma^2                      */
         omu,                /* 1 - mu                       */
         muvR,               /* mu - v_R                     */
         qfac,               /* q / sqrt(1+om^2)             */
         phase;              /* phi + acot om                */
  double *qs,                /* qfac sin(om*t[j]+phase)      */
         *e,                 /* exp(-t[j])                   */
         *omem,              /* (1-mu) (1-exp(-   t[j]))     */
         *ome2D,             /*      D (1-exp(- 2 t[j]))     */
         *pre_g,             /* g_pref /sqrt(1-exp(-2t[j]))  */
         *pre_K,             /* sqrt(t)/sqrt(1-exp(-2t[j]))  */
         eh,                 /* exp(-h/2)                    */
         omemh,              /* (1-mu) (1 - exp(-h/2))       */
         ome2Dh,             /*      D (1 - exp(-h)  )       */ 
         pre_Kh;             /* sqrt(h/2*(1-exp(-h)  )       */
} aux_type;

/***********************************************************************/
/*                                                                     */
/*                           Functions                                 */
/*                                                                     */
/***********************************************************************/

void init_calc(par_type *parp, dat_type *datp, aux_type *auxp);
void do_calc(par_type *p, dat_type *d, aux_type *a);
void save_calc(par_type *p, dat_type *d);
void clear_calc(dat_type *d, aux_type *a);
void volterra(par_type *p, dat_type *d, aux_type *a,
	      long smin, long smax, double *nrm);

void myMatPut(MATFile *fp, char *name, long rows, long cols, double *data);

/***********************************************************************/
/*                                                                     */
/*                  main program                                       */
/*                                                                     */
/***********************************************************************/

int main(int argc, char *argv[])
{
  par_type pars, *parp = &pars;    /* allocate and get pointers to data */
  dat_type dats, *datp = &dats;    /* structures                        */
  aux_type auxs, *auxp = &auxs;
  int      errflg;
  char     c;

  /* set defaults for options */
  parp->v_R      = 0.0;           /* reset potential            */
  parp->n        = -1;            /* dummy value, see init_calc */
  parp->nsec     = -1;            /* dummy value, see init_calc */
  parp->min_norm = MIN_NORM_DEF;  /* norm criterion             */
  parp->min_sec  = MIN_SEC_DEF;   /* norm criterion per section */
  parp->zip      = 0;             /* do not zip                 */

  /* process options */
  errflg = 0;
  opterr = 0;   /* errors handled by program */
  
  while (   !errflg 
         && ( c=getopt(argc,argv,":r:N:n:M:m:Zh") ) != -1 )
    switch ( c ) {
    case 'r':      /* reset potential    */
      sscanf(optarg, "%lf", &parp->v_R);
      break;
    case 'N':      /* number of points   */
      sscanf(optarg, "%ld", &parp->n);
      break;
    case 'n':      /* points per section */
      sscanf(optarg, "%ld", &parp->nsec);
      break;
    case 'M':      /* minimal norm       */
      sscanf(optarg, "%lf", &parp->min_norm);
      break;
    case 'm':      /* min. norm per sec. */ 
      sscanf(optarg, "%lf", &parp->min_sec);
      break;
    case 'Z':      /* gzip output        */
      parp->zip = 1;
      break;
    case 'h':      /* help               */
      errflg++;
      break;
    case ':':      /* missing option     */
      fprintf(stderr, "Option -%c requires an argument\n", optopt);
      errflg++;
      break;
    case '?':      /* unknown option     */
      fprintf(stderr, "Unknown option -%c\n", optopt);
      errflg++;
      break;
    default:
      fprintf(stderr, "This error is impossible\n");
      errflg++;
      break;
    }   /* switch, while */

  /* now read the required input arguments if okay */
  if ( optind + 7 > argc || errflg )
    {
      fprintf(stderr, USAGE);
      exit(-1);
    }

  sscanf(argv[optind++], "%lf", &parp->mu);
  sscanf(argv[optind++], "%lf", &parp->q);
  sscanf(argv[optind++], "%lf", &parp->om);
  sscanf(argv[optind++], "%lf", &parp->phi);
  sscanf(argv[optind++], "%lf", &parp->sigma);
  sscanf(argv[optind++], "%lf", &parp->tmax);
  parp->outname = argv[optind++];

  init_calc(parp, datp, auxp);
  do_calc(parp, datp, auxp);
  save_calc(parp, datp);
  clear_calc(datp, auxp);

  return 0;

}  /* end -- main */


/***********************************************************************/
/*                                                                     */
/*                  functions                                          */
/*                                                                     */
/***********************************************************************/

void init_calc(par_type *parp, dat_type *datp, aux_type *auxp)
{ 
  long  j;

  /* check for positive frequency */
  if ( parp->om <= 0.0 )
    {
      fprintf(stderr, "\n\n Error: Frequency must be > 0.\n\n");
      exit(-1);
    }

  /* compute number of points if not given on command line */
  if ( parp->n < 0 )
    parp->n = floor(0.5 * (double)PTS_PER_PER * parp->om * parp->tmax);

  /* check that number of points is odd and within size of AlphaBetaGamma */
  if ( !odd(parp->n) )
    parp->n++;
  if ( parp->n > N_COEFF )
    {
      fprintf(stderr, "\n\n Error: Maximum of %d points.\n\n", N_COEFF);
      exit(-1);
    }
 
  /* number of points per section if not given on command line */
  if ( parp->nsec < 0 )
    parp->nsec = parp->n - 1;  /* must be even */
  
  if ( odd(parp->nsec) )
    parp->nsec--;            

  /* make sure nsec does not exceed n */
  if ( parp->nsec >= parp->n )
    parp->nsec = parp->n - 1;

  /* convert frequency, phase to proper radians */
  parp->om  *= pi;
  parp->phi *= pi / 180;

  /* allocate memory for data and auxiliary arrays */
  datp->t     = (double *) calloc(parp->n, sizeof(double));
  datp->y     = (double *) calloc(parp->n, sizeof(double));

  auxp->qs    = (double *) calloc(parp->n, sizeof(double));
  auxp->e     = (double *) calloc(parp->n, sizeof(double));
  auxp->omem  = (double *) calloc(parp->n, sizeof(double));
  auxp->ome2D = (double *) calloc(parp->n, sizeof(double));
  auxp->pre_g = (double *) calloc(parp->n, sizeof(double));
  auxp->pre_K = (double *) calloc(parp->n, sizeof(double));

  /* verify successful allocation */
  if (   !datp->t     || !datp->y     || !auxp->qs || !auxp->e 
      || !auxp->omem  || !auxp->ome2D || !auxp->pre_g 
      || !auxp->pre_K )
    {
      fprintf(stderr,
	      "\n\n Error: Memory allocation failed (%d MB required).\n\n",
              (int) ceil(8.0*parp->n*sizeof(double)/1048576.0));
      exit(-1);
    }

  /* compute auxiliary parameters */
  auxp->h      = parp->tmax / (double) (parp->n-1);   /* time step */
  auxp->g_pref = 7.5 / sqrt(auxp->h);
  auxp->D      = parp->sigma * parp->sigma;
  auxp->omu    = 1.0 - parp->mu;
  auxp->muvR   = parp->mu - parp->v_R;
  auxp->qfac   = parp->q / sqrt(1.0 + parp->om*parp->om);
  auxp->phase  = parp->phi + atan(1.0/parp->om);

  /* fill time array */
  for ( j = 0 ; j < parp->n ; j++ )
    datp->t[j]  = (double) j * auxp->h;

  datp->y[0] = 0.0;     /* initial condition */

  /* pre-computed quantities depending on t-s only, sin, exp, 
     weight arrays, i.e. for t==0 or t==s                     
     starting points only, rest computed on the fly          */
  auxp->qs[0]    = auxp->qfac * sin(auxp->phase);
  auxp->e[0]     = 1.0;
  auxp->omem[0]  = 0.0;
  auxp->ome2D[0] = 0.0;
  auxp->pre_g[0] = 0.0;           /* by definition  */
  auxp->pre_K[0] = 1.0 / root2;   /* limiting value */
  
  /* half-interval values */
  auxp->eh     = exp(-auxp->h/2.0);
  auxp->omemh  = - auxp->omu * expm1( -auxp->h/2.0 );
  auxp->ome2Dh = - auxp->D * expm1( -auxp->h );
  auxp->pre_Kh = pow(-0.5*auxp->h/expm1(-auxp->h), 0.5);

  return;

}  /* end -- init_calc */

/***********************************************************************/

void do_calc(par_type *p, dat_type *d, aux_type *a)
{
  double nrm;       /* cumulated sum of y                */
  long   smin,      /* last point before current section */
         j;        

  nrm  = 0.0;

  smin = 0;

  /* the big loop, calculating distribution section by section */
  /* p->nsec < p->n by def. so first section fits in any case  */
  do
    {
      /* compute auxiliary parameters, weight arrays for next section */
      for ( j = smin+1 ; j <= smin + p->nsec ; j++ )
	{
	  a->qs[j]    = a->qfac * sin(p->om*d->t[j] + a->phase);
	  a->e[j]     = exp( -d->t[j] );
	  a->omem[j]  = - a->omu * expm1( -d->t[j] );
	  a->ome2D[j] = - expm1( -2*d->t[j] );
	  a->pre_g[j] = pow(a->ome2D[j], -0.5);
	  a->pre_K[j] = pow(d->t[j], 0.5) * a->pre_g[j];
	  a->pre_g[j] *= a->g_pref;
	  a->ome2D[j] *= a->D;
	}  

      /* solve next section */
      volterra(p, d, a, smin, smin+p->nsec, &nrm);

      /* iterate */
      smin += p->nsec;
    }
  while (   smin + p->nsec < p->n            /* number of points */
	 && a->h*nrm       < p->min_norm     /* total norm       */
	                                     /* relative norm    */
         && a->h*nrm       > p->min_sec  * (double) smin / (double) p->n );

  return;

}  /* end -- do_calc */

/***********************************************************************/

void save_calc(par_type *p, dat_type *d)

{
  MATFile *outfile;
  double  tmp = (double) p->nsec;
  char    outname[FILE_NAME_LENGTH],
          zip_cmd[FILE_NAME_LENGTH+20];

  /* open output file with suffix .mat */
  sprintf(outname, "%s.mat", p->outname);
  outfile = matOpen(outname, "w");
  if ( !outfile )
    {
      fprintf(stderr, "\nCould not matOpen %s !\n", p->outname);
      exit(-1);
    }

  /* put FPT density, time axis, and parameters into file */
  myMatPut(outfile, "rho",     p->n, 1,   d->y     );
  myMatPut(outfile, "dt",         1, 1, &(d->t[1]) );
  myMatPut(outfile, "t",       p->n, 1,   d->t     );   
  myMatPut(outfile, "mu",         1, 1,  &p->mu    );
  myMatPut(outfile, "q",          1, 1,  &p->q     );
  myMatPut(outfile, "om",         1, 1,  &p->om    );
  myMatPut(outfile, "phi",        1, 1,  &p->phi   );
  myMatPut(outfile, "sigma",      1, 1,  &p->sigma );
  myMatPut(outfile, "v_R",        1, 1,  &p->v_R   );
  myMatPut(outfile, "nsec",       1, 1,  &tmp      );

  matClose(outfile);

  /* compress */
  if ( p->zip )
    {
      sprintf(zip_cmd, "gzip -f %s", outname);
      system(zip_cmd);
    }

  return;

}  /* end -- save_calc */

/***********************************************************************/

void myMatPut(MATFile *fp, char *name, long rows, long cols, double *data)

{
  mxArray *mdata;

  /* see MATLAB External Interface Handbook */
  mdata = mxCreateDoubleMatrix(rows, cols, mxREAL);
  memcpy(mxGetPr(mdata), data, rows*cols*sizeof(double));
  mxSetName(mdata, name);
  matPutArray(fp, mdata);
  mxDestroyArray(mdata);

  return;
}  /* end -- myMatPut */

/***********************************************************************/

void clear_calc(dat_type *datp, aux_type *auxp)

{
  free(datp->t    );
  free(datp->y    );
  free(auxp->qs   );
  free(auxp->e    );
  free(auxp->omem );
  free(auxp->ome2D);
  free(auxp->pre_g);
  free(auxp->pre_K);
  return;

}  /* end -- clear_calc */

/***********************************************************************/

void volterra(par_type *p, dat_type *d, aux_type *a,
	      long smin, long smax, double *nrm)
{
  /* this function computes y(t) for t[smin+1]..t[smax] given
     y[0]..y[smin]                                            */
 
  long int  m, i;
  double    S1, S2, P, Q, R, T, V, W, G;

  /* outer loop -- points of desired function                               */
  /* NOTE: use m < smax-1 here to be on the save side even when smax is odd */
  /* NOTE: the m used here is 2m in Plesser, 1999                           */
  for ( m = smin ; m < smax-1 ; m += 2 )
    {
      /* the various kernel evaluations */
      P =            /* K(t[m+2], t[m]) */
        K(a->pre_K[2], a->e[2], a->omem[2], a->ome2D[2], a->qs[m+2], a->qs[m]);
      Q =            /* K(t[m+1], t[m+1/2]) */
        K(a->pre_Kh, a->eh, a->omemh, a->ome2Dh, a->qs[m+1], 
                        a->qfac * sin(p->om*(d->t[m]+0.5*a->h)+a->phase));
      R =            /* K(t[m+1], t[m]) */
        K(a->pre_K[1], a->e[1], a->omem[1], a->ome2D[1], a->qs[m+1], a->qs[m]);
      T =            /* K(t[m+2], t[m+1]) */
        K(a->pre_K[1], a->e[1], a->omem[1], a->ome2D[1],a->qs[m+2],a->qs[m+1]);
      V =            /* K(t[m+1], t[m+1]) */ 
        1.0/root2;  
      W =            /* K(t[m+2], t[m+2]) */ 
        1.0/root2;  

      /* determinant of iteration matrix */
      G = W*(Q+V)+Q*T/4.5;

      /* "fix" terms of right hand side */
      S1 = 
      g(a->pre_g[m+1],a->e[m+1],a->ome2D[m+1],a->qs[m+1],a->qs[0],
                                                          a->omu, a->muvR)
           - (R+3.0*Q) * d->y[m];
      S2 =
      g(a->pre_g[m+2],a->e[m+2],a->ome2D[m+2],a->qs[m+2],a->qs[0],
                                                          a->omu, a->muvR)
           - root2 * P * d->y[m];

      /* inner loop -- integral over known points */
      if ( m > 0 )
	{
	  S1 -=  Alpha[m+1] * d->y[0] * K(a->pre_K[m+1],a->e[m+1],a->omem[m+1],
                                          a->ome2D[m+1],a->qs[m+1],a->qs[0])
                +Beta[m]    * d->y[1] * K(a->pre_K[m  ],a->e[m  ],a->omem[m  ],
                                          a->ome2D[m  ],a->qs[m+1],a->qs[1])
                +Gamma[1]   * d->y[m] * R;

	  S2 -=  Alpha[m+2] * d->y[0] * K(a->pre_K[m+2],a->e[m+2],a->omem[m+2],
                                          a->ome2D[m+2],a->qs[m+2],a->qs[0])
                +Beta[m+1]  * d->y[1] * K(a->pre_K[m+1],a->e[m+1],a->omem[m+1],
                                          a->ome2D[m+1],a->qs[m+2],a->qs[1])
                +Gamma[2]   * d->y[m] * P;

	  for ( i = 2 ; i < m ; i += 2 )
	    {
	      S1 -= AlphaGamma[m+1-i] * d->y[i]   *             /* even */
                       K(a->pre_K[m+1-i],a->e[m+1-i],a->omem[m+1-i],
                         a->ome2D[m+1-i],a->qs[m+1],a->qs[i]);
	      S1 -= Beta[m-i]         * d->y[i+1] *             /* odd  */
                       K(a->pre_K[m  -i],a->e[m  -i],a->omem[m  -i],
                         a->ome2D[m  -i],a->qs[m+1],a->qs[i+1]);
	      
	      S2 -= AlphaGamma[m+2-i] * d->y[i]   *             /* even */
                       K(a->pre_K[m+2-i],a->e[m+2-i],a->omem[m+2-i],
                         a->ome2D[m+2-i],a->qs[m+2],a->qs[i]);
	      S2 -= Beta[m+1-i]       * d->y[i+1] *             /* odd  */
                       K(a->pre_K[m+1-i],a->e[m+1-i],a->omem[m+1-i],
                         a->ome2D[m+1-i],a->qs[m+2],a->qs[i+1]);
	    }
	}  /* end -- if m > 0 */

      /* compute the next two points */
      d->y[m+1] = ( W*S1/6.0   + Q    *S2/36.0/root2 ) / G;
      d->y[m+2] = (-T*S1/4.5   + (Q+V)*S2/ 6.0/root2 ) / G;

      /* stabilize by keeping non-negative, accumulate "norm" */
      if ( d->y[m+1] < 0.0 ) d->y[m+1] = 0.0;
      if ( d->y[m+2] < 0.0 ) d->y[m+2] = 0.0;

      *nrm += d->y[m+1]+d->y[m+2];

    }

  return;

}  /* end -- volterra */



      
      
