function [S, omS] = S_approx(t, isi, cisi, To, nperpsi, nfft, nspk),

%  Compute approximation to full power spectrum of spike train
%
%    [S, omS] = S_approx(t, isi, cisi, To, nperpsi [, nfft [, nspk]])
%
%  Input:
%    isi(t)     Stationary ISI density
%    cisi(t,:)  Conditional ISI densities
%    To         observation time
%    nperpsi    isi points per psi, same as in ou_markov call
%    nfft       2^nfft is FFT length to use [ shortest possible ]
%    nspk       compute PSD for nspk spikes [ floor(To/<tau>) ]
%
%  Output:
%    S(omS)     power spectrum

%
%  Literature: Hans E. Plesser, Dissertation, 1999, Chs. 4.6.3, 4.4.3
%

%  Copyright (C) 2000 Hans Ekkehard Plesser
%
%  This program is free software; you can redistribute it and/or
%  modify it under the terms of the GNU General Public License
%  as published by the Free Software Foundation; either version 2
%  of the License, or (at your option) any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program; if not, write to the Free Software
%  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


% mean ISI
tau = trapz(t, t.*isi);

% number of spikes to compute PSD from, if not given as input
if nargin < 7,
  nspk = floor(To/tau);
end

% time limit for FFT
tmax = To;   

% ensure we got a train
if nspk < 2,
  error('No spectrum for less than two spikes!');
end

% determine size of cISI and transition matrix
[npts, npsi] = size(cisi);
nper = npts / (nperpsi*npsi);    % number of periods in cisi
nQ   = nper * npsi;              % "length" of matrix Q

% build transition matrix Q, Eq. (4.31)
% - Q is lower triangular matrix
% - built of nper "hypercolumns" of npsi columns each
% - leftmost hypercolum contains one column for each psi
%   * first column is cisi(:, 1)
%   * subsequent columns are subsequent columns of cisi, each
%     shifted down one row and truncated at bottom
%   * NOTE: cisi(1,:) == 0 because of initial cond, so diagonal 0
% - each hypercolumn is a copy of its left neighbor, shifted 
%     down by npsi rows, truncated at bottom
% - time discretization discretization is coarsened to phase
%   discretization, i.e., omega dt = dpsi, to have same time
%   discretization along rows and columns
dt = diff(t([1 nperpsi+1]));
Q  = zeros(nQ);
ix = 1:nperpsi:npts;             % coarsening index
tQ = t(ix);                      % time axis for Q
for p = 1:nper,
  for j = 1:npsi,
    c = (p-1)*npsi + j;          % column number & row of top entry
    Q(c:nQ, c) = cisi(ix(1:(nQ-c+1)), j);
  end
end
Q = dt * Q;                      % scale by dt


% reduce transition matrix to within To
nq = min(find(tQ>=tmax));             % points within time-limit
if isempty(nq), nq = length(tQ); end  % handle rounding problems
Q  = Q(1:nq, 1:nq);

% determine FFT length
if nargin < 6,             % FFT length given as input
  nfft = 2^ceil(log2(nq));
elseif nfft < nq,          % given FFT length too short
  nfft = 2^ceil(log2(nq));
end


%
%  Evaluation of Eq. (4.66)
%

% do things as compact as possible to save memory
Qp  = eye(nq);           % current power of Q, start at Q^0
Qp1 = zeros(nq-1, nq);   % first column of Qp as rows
Qs  = zeros(nq, nq);     % sum of powers

% compute individual terms in the sum 
%    \{ \sum_{j=1}^{M-1} [ ... ] \hat{Q}_{l1}^j \}
% - Qp1 collects \hat{Q}_{l1}^j as rows
% - Qf{j} collects [ ... ] 
for j = 1:nspk-1,
  Qp = Q * Qp;             % next power of Q
  
  % renormalize each column of Qp as distribution
  Qpsc = Qp;               % copy
  qnrm = sum(Qpsc);        % Q was norm'ed as distribution, not density
  ix   = find(qnrm > 0);   % do not renorm zeros
  Qpsc(:,ix) = Qpsc(:,ix) ./ repmat(qnrm(ix), [nq, 1]);  % zeros not norm'ed

  % collect first column as rows for sum_{l=1}^N ...
  Qp1(j, :) = Qpsc(:, 1).';

  % add next term to sum of powers, i.e., ( sum_{k=1}^{M-j} \hat{Q}_{nl}^k )
  Qs = Qs + Qpsc;
  
  % [ sum_{n=l+1}^L e^{...} (...) ] 
  % - evaluated by columnwise FFT
  % - exploit that Q is lower triangular
  % - keep only first half of FFT because of symmetry
  Qff = fft(Qs, nfft);
  Qf{nspk-j} = Qff(1:nfft/2, :);
end

% summation over j, i.e. \{ \sum_{j=1}^{M-1} [ ... ] \hat{Q}_{l1}^j \}
%   resulting matrix has one row for each point of spectrum
Qp = zeros(size(Qf{1}));    % re-use Qp to save memory
for j = 1:nspk-1, 
  Qp = Qp + Qf{j}.* repmat(Qp1(j,:), [nfft/2, 1]);
end

omS  = (0:nfft/2-1).'*2*pi/(nfft*dt);    % frequency axis
sc   = zeros(nfft/2, 1);                 % spectrum will be collected here
tfft = reshape(tQ(1:nq),nq,1);           % time acis  

% sum_{l=1}^N \{ ... \}
% - evalutate row by row for each point of spectrum
for j = 1:nfft/2,
  sc(j) = Qp(j, :) * exp(i*omS(j)*tfft); % row * column = scalar
end

% add Poisson and scaling factor for full spectrum
S = 1/(pi*tau) + 2/(pi*tau*nspk) * real(sc);

return

