function [S, contS] = ...
                  S_harm(t, isi, psi, chi, TM, ev, C, iC, T_o, harm, M),

%  Compute the PSD at the stimulus harmonics via phase Markov chain
%
%    [S, contS] = 
%          S_harm(t, isi, psi, chi, TM, ev, C, iC, T_o [, harm [, M]])
%
%  Input:
%    t, isi     isi(t) is stationary ISI density
%    psi, chi   chi(psi) is stationary phase density
%    TM         TM(psi, psi) is phase transition matrix
%    ev         vector of TM's eigenvalues, must have ev(1)==1, |ev(j)| < 1
%    C, iC      iC * TM * C = diag(ev)
%    T_o        observation time
%    harm       harmonics to be evaluated [ def 1, may be vector ]
%    M          number of spikes to compute spectrum for [ def. T_o/<tau> ]
%
%  Output:
%    S          PSD at harmonics
%    contS      continuous part of spectrum

%
%  Literature: Hans E. Plesser, Dissertation, 1999, Ch. 4.6.2
%

%  Copyright (C) 2000 Hans Ekkehard Plesser
%
%  This program is free software; you can redistribute it and/or
%  modify it under the terms of the GNU General Public License
%  as published by the Free Software Foundation; either version 2
%  of the License, or (at your option) any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program; if not, write to the Free Software
%  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


% check for list of harmonics
if nargin < 10,
  harm = 1;
end

% check for spike number
tau = trapz(t, t.*isi);   % mean ISI
if nargin < 11,
  M = T_o / tau;
end

%  ensure correct orientation of vectors
chi  = chi(:);
psi  = psi(:);
harm = harm(:).';
nharm= length(harm);

% side vectors, Eq. (4.54) (matrices, so we can handle harmonics in parallel)
a = exp(i*psi*harm);
b = (chi * ones(1, nharm)) .* conj(a);

% singular part of spectrum  --- independent of M, Eq. (4.61)
% use explicit formulae to avoid numerical problems in C, iC
Bn = abs(sum(repmat(chi,1,nharm).*a)).^2;

% continuous part, infinite term, cf. Eqs. (4.60), (4.56) w. m>1, first term
ev(1)   = 0;              % is handeled by Bn
TD_inf  = C * diag(ev ./ (1-ev)) * iC;
AnInf   = 2 * real(sum( a.* (TD_inf * b) ));

% continuous part, finite term, cf. Eqs. (4.60), (4.56) w. m>1, second term
TD_M    = C * diag(ev .* (ev.^M-1) .* (ev-1).^(-2)) * iC;
AnM     = 2/M * real(sum( a.* (TD_M * b) ));

% build spectrum, return column vectors, Eq. (4.62)
S    = ( 1 + AnInf(:) + AnM(:) + (M-1)*Bn(:) ) / ( pi*tau );

% build background
if nargout > 1,
  contS = ( 1 + AnInf(:) + AnM(:) ) / ( pi * tau );
end

% if requested, add infinite amplitudes
if nargout > 2,
  A_inf = AnInf(:) / (pi*tau);
  B_inf = Bn(:) / (pi*tau);
end

return
