function [t, isi, psi, chi_s, TM, cisi, ev, C, iC, props] = ...
    ou_markov(mu, q, Om, sigma, v_res, npsi, tmax, nperpsi),

%
%  Markov analysis of the periodically driven OU neuron
%
%  [t, isi, psi, chi_s, TM, cisi, ev, C, iC, props] = ...
%           ou_markov(mu, q, Om, sigma, v_res, npsi, tmax, nperpsi)
%  Input:
%        mu  DC input                                         
%         q  amplitude of AC input			       
%        Om  stimulus frequency / pi [ in radians ]
%     sigma  rms amplitude of noise			       
%     v_res  reset potential
%      npsi  phase axis discretized to npsi points [ must be even ]
%      tmax  time limit for computation		       
%   nperpsi  no. of points for ISI computation per stimulus period
%              and per phase discretization point 
%              [ must be even, typical 2, 4, 8 ]
%
%  Output:
%       isi(t)  Stationary interspike-interval density, Eq. (4.26)
%   chi_s(psi)  Stationary spike phase density, Eqs. (4.23), (4.30)
%  TM(psi,psi)  Phase transition matrix, Eqs. (4.18), (4.28)
%  cisi(t,psi)  Conditional ISI density, Eq. (4.15)
%           ev  Eigenvalues of iteration Eq. (4.29), sorted
%        C, iC  Transformation matrix, inverse, Eq. (4.57)
%        props  Data structure with quantities for control:
%                 nrm_cisi         norm of conditional isi's
%                 rk_TM, rk_C      ranks of TM and C
%                 imag_chi         norm of imaginary part of C(:,1) ~ chi_s
%                 chi_min          min. of chi_s prior to elimination of negs
%                 rcond            estimated condition of C
%                 resid            norm(C*diag(ev)*iC - TM)
%                 resid_inv        norm(C*iC - eye())
%
%  Notes:
%    * Computation over Nper = ceil(Om tmax / 2pi) full periods, 
%        i.e., effective Tmax = 2/Om * Nperiods >= tmax
%    * Nper must be even
%    * nperpsi * npsi ISI points per stimulus period, must be even

%  Requires p_fptd as MEX function from ModUhl/fptd/MEX
%
%  Literature: Hans E. Plesser, Dissertation, 1999, Ch. 4.4

%  Copyright (C) 2000 Hans Ekkehard Plesser
%
%  This program is free software; you can redistribute it and/or
%  modify it under the terms of the GNU General Public License
%  as published by the Free Software Foundation; either version 2
%  of the License, or (at your option) any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program; if not, write to the Free Software
%  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

% check that Om is positive
if Om <= 0,
  error(' Frequency must be positive! ');
end

% check that sigma is positive
if sigma <= 0,
  error(' Noise amplitude must be positive! ');
end

% make nperpsi even
nperpsi = nperpsi + rem(nperpsi, 2);

% phase axis
psi    = -pi + (0:npsi-1).' / npsi * 2 * pi;
psideg = psi / pi * 180;      

% time axis---build before calls to p_fptd, so we can get array sizes, etc
[nt, t_rho, dt, Nper, nper, nsec] = time_axis(Om, nperpsi, npsi, tmax);
t = t_rho(1:nt);     % discard dummy point

% build phase transition matrices, conditional ISI density
TM   = repmat(NaN, [npsi, npsi]);  % phase transition matrices
cisi = repmat(NaN, [nt,   npsi]);  % conditional ISI

% build props structure
props.nrm_cisi = repmat(NaN, [npsi, 1]);
props.rk_TM    = NaN;
props.rk_C     = NaN;
props.imag_chi = NaN;
props.chi_min  = NaN;
props.rcond    = NaN;
props.resid    = NaN;
props.resid_inv= NaN;

% set up rotation of cISIs for evaluation of Eq. (4.28):
%   cisi columns are rotated by dpsi/2=nperpsi/2 so that values are
%   properly centered in bins
hdpsi   = nperpsi / 2;
rot_idx = [(nper-hdpsi+1):nper 1:(nper-hdpsi)].';

% compute conditional ISI's and compile
for k = 1:npsi,

  rho = p_fptd(mu, q, Om, psideg(k), sigma, t(end), 'r', v_res, ...
                'N', nt+1, 'n', nsec, 'M', 0.999, 'm', 0.9);

  % terminate quickly in hopeless cases
  if trapz(t_rho, rho) < 0.8,
    isi   = repmat(NaN, (size(t)));
    chi_s = repmat(NaN, (size(psi)));
    props.nrm_cisi = trapz(t, cisi).';
    if nargout > 6, ev   = repmat(NaN, [npsi,    1]); end
    if nargout > 7,  C   = repmat(NaN, [npsi, npsi]); end
    if nargout > 8, 
         iC     = repmat(NaN, [npsi, npsi]); 
         cond_C = NaN;
         res_C  = NaN;
       end
    return
  end
	    
  % discard dummy point at end
  cisi(:, k) = rho(1:nt);     
  
  % block and sum cisi by period---sum in Eq. (4.28)
  psum = sum(reshape(cisi(:,k), nper, Nper), 2);
  
  % Evaluation of integral in Eq. (4.28)
  % 1. add neighboring points on fine phase axis (nperpsi*npsi) to
  %    imitate trapezoidal rule; use dummy point at end
  psum = psum + [psum(2:nper) ; psum(1)-rho(1)+rho(nt+1)];
  
  % 2. rotate by half dpsi to get psi0 into centers of integration intervals
  psum = psum(rot_idx);
  
  % 3. complete integration by summation and scaling by 0.5*dt
  %    (factor 0.5 is for trapeziodal rule)
  % 4. assign into properly rotated transition matrix
  TM([k:npsi 1:k-1], k) = 0.5 * dt * sum(reshape(psum, nperpsi, npsi)).';
  
end

% re-normalize TM to correct for numerical losses in norm
nrm = sum(TM);                       % nrm > 0, see check above
TM  = TM ./ repmat(nrm, [npsi, 1]);

% eliminate numerical noise
max_TM = max(TM(:));
TM(TM < eps*max_TM) = 0;

% properties of ISIs, TM
props.nrm_cisi = trapz(t, cisi).';
props.rk_TM    = rank(TM);

% compute eigenvalues, -vectors, guard for numerical problems
try
  [C, D] = eig(TM);
catch
  try      % sometimes the shifted matrix converges better
    TM = fftshift(TM);
    [C, D] = eig(TM);

    % to get correct phase order back, swap upper and lower halfs
    C = [C((npsi/2+1):npsi,:) ; C(1:npsi/2,:)];
  catch
    warning('Cannot compute eigenvalues');
    isi   = repmat(NaN, [nt,   1]); 
    chi_s = repmat(NaN, [npsi, 1]);
    if nargout > 6, ev = repmat(NaN, [npsi, 1]); end
    if nargout > 7,  C = repmat(NaN, [npsi, npsi]); end
    if nargout > 8, 
      iC     = repmat(NaN, [npsi, npsi]); 
      cond_C = NaN;
      res_C  = NaN;
    end
    return
  end
end

% extract eigenvalues from matrix D, sort in descending order of abs
ev        = diag(D);
[ews, ix] = sort(-abs(ev));  % descending order of absolute values
ev        = ev(ix);
C         = C(:, ix);        % sort eigenvectors correspondingly

% properties of C, ev
props.rk_C   = rank(C);
props.rcond  = rcond(C);

% stationary state from largest ev
chi_s          = real(C(:, 1));

% renormalize chi_s, removes unwanted factor of -1 if around
nrm_chi        = sum(chi_s);
chi_s          = chi_s / nrm_chi;

% properties of chi_s, eliminate negatives
props.imag_chi   = norm(imag(C(:,1)));
props.chi_min    = min(chi_s);
chi_s(chi_s < 0) = 0;

% compute ISI density
isi = cisi * chi_s;

% inverse trafo matrix if wanted, get residue
% inversion by \ reduces residue by factor of 10, but sometimes
%   bounces where inv works.  Thus try both.
if nargout > 8,

  iC=(C*diag(ev)) \ TM; 

  if isinf(iC(1)),
    try 
      iC = inv(C);
    catch
      iC = repmat(NaN, size(C));
      return    
    end
  end
  
  % test inverse and eigenvalues
  props.resid     = norm(C*diag(ev)*iC-TM);
  props.resid_inv = norm(C*iC-eye(size(C)));

end

return



%  function computing time axis and corresponding params
function [nt, t, dt, Nper, nper, nsec, T] = time_axis(Om, nperpsi, npsi, tmax),

T    = 2 / Om;            % period length
Nper = ceil(tmax / T);    % number of periods, >= 1
Nper = Nper + rem(Nper,2);% must be even for splitting in secs of 2*nper
nper = nperpsi * npsi;    % number of points per period
nsec = 2*nper;            % number of points per section
nt   = Nper * nper;       % total number of points
nt   = nt - rem(nt, 2);   % make even, dummy point added at end for oddity
dt   = (Nper*T) / nt;     % time discretization
t    = (0:nt).' * dt;     % dummy added here; only t(1:nt) are used
                          % in analysis

% avoid overly long sections for low frequencies
% overly short for high frequencies---this is brute force!
if nsec > nt,
  nsec = nt;
elseif nsec > nt / 8 & nsec > 1024, 
  if rem(nt, 8) == 0,
    nsec = nt / 8;
  else
    nsec = nt;
  end
end


return
