function [q, tq, Q, qk, F, pk, pkk] = ...
                        spike_intensity(t, cisi, nperpsi, T_o, kmax),

%   Compute spike intensity of periodically driven OU neuron
%
%    [q, tq , Q, qk, F, pk, pkk] = 
%              spike_intensity(t, cisi, nperpsi [, T_o [,kmax]])
%
%   Input:
%     isi(t)     Stationary ISI density
%     cisi(t,:)  Conditional ISI densities
%     nperpsi    isi points per psi, same as in ou_markov call
%     T_o        observation time, needed for pk, F only
%     kmax       compute densities for spikes 1..kmax [ def. 20 ]
%
%   Output:
%     q(tq)      Spike intensity
%     Q          transition matrix
%     qk(tq,:)   Intensity for spike 1..kmax
%     F          Fano factor for obs. time T_o
%     pk(pkk)    Prob to have pkk spikes in 0..T_o
%
%   NOTE: 
%   - All return arguments except q, tq are "optional" and computation
%     time is markedly shorter if function is invoked with only two
%     return arguments.
%   - Calculation of the Fano factor is approximate only.  Check the
%     spike number distribution pk for sanity!
%   - Computation of q(tq) is fast, while computation of qk, F, pk can
%     be EXTREMELY slow, as it requires kmax multiplications of matrices
%     of size (rows(cisi)/nperpsi)^2.  You should use small npsi, tmax,
%     and rather large nperpsi in the ou_markov call when intending to
%     compute qk, F, pk.  E.g., ou_markov(mu,q,Om,sigma,vres,36,200,8).
%

%  Literature: Hans E. Plesser, Dissertation, Ch. 4.4.3, 4.5
%

%  Copyright (C) 2000 Hans Ekkehard Plesser
%
%  This program is free software; you can redistribute it and/or
%  modify it under the terms of the GNU General Public License
%  as published by the Free Software Foundation; either version 2
%  of the License, or (at your option) any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program; if not, write to the Free Software
%  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


% reduce cisi to square matrix per period by coarsening time scale
[npts, npsi] = size(cisi);

nper = npts / (nperpsi*npsi);    % number of periods in cisi

tq = t(1:nperpsi:end);
dt = diff(tq(1:2));
N  = length(tq);
nQ = nper * npsi;

if nQ ~= N,
  error('sizes incompatible')
end

% iteratively compute spike intensity, Eq. (4.35)
q    = zeros(N, 1);
q(1) = 1/dt;         % ensures normalization as density for all

for m = 2:length(tq),
  % Mathematical indeces running from 0..N-1, 0..npsi-1
  % q_m  = sum(Q_{m,j}*q_j, j=0..m-1) 
  %      = sum(cISI_{m-j, j mod npsi}*q_j, j=0..m-1) 
  % MATLAB indeces, running from 1..N, 1..npsi
  % q(m) = sum(cisi(m-j+1, rem(j-1, npsi)+1)*q(j), j=1:m-1)

  % evaluate with direct indexing, use only every nperpsi-th row
  j  = 1:m-1;
  ix = 1 + (m-j) * nperpsi + rem(j-1, npsi) * npts; 
  
  % multiply with dt to obtain proper elements of Q
  q(m) = dt * cisi(ix) * q(1:m-1);
end

% eliminate spike at t=0, matter of definition
q(1) = 0;

% check if user wants anything more
if nargout < 3,
  return
end

% build transition matrix Q, Eq. (4.31)
% - Q is lower triangular matrix
% - built of nper "hypercolumns" of npsi columns each
% - leftmost hypercolum contains one column for each psi
%   * first column is cisi(:, 1)
%   * subsequent columns are subsequent columns of cisi, each
%     shifted down one row and truncated at bottom
%   * NOTE: cisi(1,:) == 0 because of initial cond, so diagonal 0
% - each hypercolumn is a copy of its left neighbor, shifted 
%     down by npsi rows, truncated at bottom
% - time discretization discretization is coarsened to phase
%   discretization, i.e., omega dt = dpsi, to have same time
%   discretization along rows and columns
Q  = zeros(nQ);
for p = 1:nper,
  for j = 1:npsi,
    c = (p-1)*npsi + j;  % column number, first entry
    ce= 1 + (nQ-c)*nperpsi; % last to take from cisi
    Q(c:nQ, c) = cisi(1:nperpsi:ce, j);
  end
end

% scale by dt to obtain discrete distribution
Q = dt * Q;

% wants only Q?
if nargout < 4,
  return
end

% compute densities for 1..kth spike, Eq. (4.33)
if nargin < 5,
  kmax = 20;    % sanity limit
end

Qk = Q;
for j = 1:kmax,

  % due to initial condition, density of k-th spike is just first
  % column of k-th power of matrix Q
  qk(:, j) = Qk(:, 1);
  
  % check norm, stop if we are just iterating zeros
  if sum(qk(:, j)) < 0.0001,
    break;
  end
  
  Qk = Q * Qk;    % iterate
  
end

% convert to density
qk = qk / dt;

% probability of finding k spikes, Fano factor
nk       = size(qk, 2);
pkk      = 0:nk-1;
ix_To    = min(find(T_o <= tq));

% check if T_o in time range
if isempty(ix_To),
  error('Fano factor calculation requires T_o < max(t).')
end
pk(1)    = max(0, 1 - trapz(tq(1:ix_To), qk(1:ix_To,1)));
pk(2:nk) = trapz(tq(1:ix_To), qk(1:ix_To,1:end-1)-qk(1:ix_To,2:end)).';

km = sum(pkk.*pk);
if km > 0,
  F = (sum(pkk.^2.*pk)-km^2)/km;
else
  F = NaN;
end


return


