function [S, om] = renewal_spectrum(rho, tau, log2_n),

%    [S, om] = renewal_spectrum(rho, tau, log2_n)
%
%  Calculate the power spectral density of a renewal process,
%  see Ch. 4.6.1 of Plesser (1999)
%
%  rho, tau   rho(tau) is interval density of the process, 
%             from p_fptd or ap_fptd
%             tau must start at 0 and be equispaced
%
%  log2_n     spectrum is calculated at 2^(log2_n-1)+1 points, 
%             2^log2_n > length(tau)
%
%  S, om      S(om) is the power spectral density
%
%  if rho is a matrix, the PSD calculated per column

%  Copyright (C) 2000 Hans Ekkehard Plesser
%
%  This program is free software; you can redistribute it and/or
%  modify it under the terms of the GNU General Public License
%  as published by the Free Software Foundation; either version 2
%  of the License, or (at your option) any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program; if not, write to the Free Software
%  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

%  for theory see
%
%   Franklin & Bair, SIAM J Appl Math 55:1074 (1995)

% get length of spectrum, half length
m = 2^log2_n;
mh = 2^(log2_n-1);

% insure column vector tau, get time spacing, frequency axis
tau = tau(:);
dt  = tau(2);
if ( nargout > 1 )
  om = 2*pi/m/dt * (0:mh).';
end

% size of density matrix
[rows, cols] = size(rho);

% check arguments
if ( length(tau) ~= rows )
  error('rho must have as many rows as tau has entries.')
elseif ( length(tau) > m )
  error('FFT length is shorter than tau.')
elseif ( tau(1) ~= 0 )
  error('tau(1) ~= 0 ')
elseif ( std(diff(tau))/mean(diff(tau)) > 1e-7 )
  error('tau must have fixed spacing')
end

% calculate mean and variance of interval length, coeff of variation
mu  = trapz(tau, (tau    * ones(1,cols)) .* rho );
var = trapz(tau, (tau.^2 * ones(1,cols)) .* rho ) - mu.^2;
Cv  = sqrt(var) ./ mu;

% calculate FFT
% zero padding to length m is automatic
% scaling is not built in, so do by hand
h = dt * fft(rho, m);

% calculate transform of renewal density, 
% need first half only, om = 0 is handled via limit below
ix = 2:(mh+1);
G  = 2 * real(h(ix,:) ./ ( 1 - h(ix,:) ));

% calculate spectrum
S(ix, :) = 1/(pi*mu) * ( 1 + G );
S( 1, :) = Cv.^2 / pi / mu;       

return

