/*

  ap_train.c   Simulate spike train in response to aperiodic stimulus
                 Hyperpolarization can be limited
                 Reset potential can be different from resting potential
  	         Stimulus is read from file, all file I/O with MATLAB files
	         If desired, a file containing a voltage trace is saved
		 All file I/O with MATLAB files
		 Stimulus file from ap_stim.m

    	       Literature: Hans E. Plesser, Dissertation (1999), Appendix B
                           D. Gillespie, Phys Rev E 54:2084-2091 (1996)


  Copyright (C) 2000 Hans Ekkehard Plesser

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  The calling syntax is:
*/


#define USAGE \
"                                                                      \n"\
" Simulation of spike trains for aperiodic stimuli                     \n"\
"                                                                      \n"\
"  ap_train [options] mu q rms stimulus outfile                        \n"\
"                                                                      \n"\
"  Input:         mu  mean value of input                              \n"\
"                  q  rms amplitude of input			       \n"\
"                rms  rms amplitude of noise			       \n"\
"           stimulus  MATLAB file containing dt, F_n from ap_stim.m    \n"\
"            outfile  MATLAB file containing n_spike t_spike	       \n"\
"								       \n"\
"  Options:   -d ###  minimum step size [ 1e-9 ]                       \n"\
"             -y ###  minimum potential [unbounded hyperpolarization ] \n"\
"             -r ###  reset potential   [ 0    ]		       \n"\
"             -R      reset time after each spike [ no reset ]         \n"\
"             -M ###  max. number of spikes [ 10000 ]		       \n"\
"             -T xxx  write potential trace to file xxx   [ no trace ] \n"\
"             -s ###  random generator seed [ use clock ]              \n"\
"                                                                      \n"\
"  Notes:     * Maximum step size is stimulus time resolution.         \n"\
"             * For finer step size, stimulus is linearly interpolated.\n"\
"                                                                      \n"\
"             Constrained stimulation:                                 \n"\
"             * -R forces time reset to t = 0 (abs refr time) after    \n"\
"               each spike.                                            \n"\
"             * ISI cannot exceed length of stimulus provided.         \n"\
"             * The output file contains number of spikes in n_spike   \n"\
"               and interspike-intervals in t_spike.                   \n"\
"                                                                      \n"\
"             Unconstrained stimulation:                               \n"\
"             * Spike train cannot exceed length of stimulus provided. \n"\
"             * The output file contains number of spikes in n_spike   \n"\
"               and absolute spike times in t_spike.                   \n"\
"                                                                      \n"\
"             Voltage trace:                                           \n"\
"             * Option -T is EXTREMELY memory-hungry, since the trace  \n"\
"               is buffered in memory during simulation. Use with care!\n"\
"             * Trace file contains number of trace points in n_trace, \n"\
"               and v_trace(t_trace) is voltage trace                  \n"\
"                                                                      \n"


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <string.h>
#include <getopt.h>

#include <rand55.h>   /* random number generator */

#include <mat.h>      /* MATLAB matrix handling  */

/***********************************************************************/
/*                                                                     */
/*                  Macros                                             */
/*                                                                     */
/***********************************************************************/

#define	max(A, B)	((A) > (B) ? (A) : (B))
#define	min(A, B)	((A) < (B) ? (A) : (B))
#define odd(n)          ( (n) & 1 )

#define pi     3.14159265358979324
#define root2  1.41421356237309505

#define THRESHOLD_DISTANCE 5.0   /* for choice of dt, Eq. (B.4) */

#define FILE_NAME_LENGTH   255

/***********************************************************************/
/*                                                                     */
/*                  data structures                                    */
/*                                                                     */
/***********************************************************************/

typedef struct              /* input/problem parameters      */
{
  long   n_in,              /* number of pts in stimulus     */
         max_spikes,        /* maximum number of spikes      */
         seed;              /* random generator seed         */
  int    reset_t,           /* reset to t=0 after each spike */
         do_trace;          /* record membrane pot. trace    */
  double dt_in,             /* time step of input signal     */
         dt_min,            /* minimum time step             */
         dt_trace,          /* minimum step for volt. tr.    */
         *iS,               /* integrated input signal       */
         mean,              /* stimulus mean                 */
         sigma,             /* stimulus amplitude            */
         rms,               /* noise amplitude               */
         v_hyp,             /* minimum hyperpolarization     */
         v_res;             /* reset potential               */
  char   *inname,           /* input file name               */
         *outname,          /* name of output file           */
         *tracename;        /* name of voltage trace file    */
} par_type;

typedef struct              /* spike train, volt. trace      */
{
  long    n_spike,          /* number of spikes so far       */
          n_trace;          /* number of trace entries       */
  double *t_spike,          /* spike times                   */
         *t_trace,          /* time axis of voltage trace    */
         *v_trace;          /* voltage trace                 */
} dat_type;

typedef struct              /* auxiliary quantities          */
{
  long    n_max;            /* length of trace array         */
  double  t_max,            /* end of stimulus               */
          v_max,            /* dt_in  for v < v_max          */
          v_min,            /* dt_min for v > v_min          */
          sig_min,          /* corrsponding variances        */
          sig_max;
} aux_type;

/***********************************************************************/
/*                                                                     */
/*                           Functions                                 */
/*                                                                     */
/***********************************************************************/

void init_calc(par_type *parp, dat_type *datp, aux_type *auxp);
void do_calc(par_type *p, dat_type *d, aux_type *a);
void save_calc(par_type *p, dat_type *d);
void clear_calc(par_type *p, dat_type *d, aux_type *a);

double next_spike(par_type *p, dat_type *d, aux_type *a);

/* simplify MATLAB matrix handling */
long myMatGet(MATFile *fp, const char *name, double **data);
void myMatPut(MATFile *fp, char *name, long rows, long cols, double *data);


/***********************************************************************/
/*                                                                     */
/*                  main program                                       */
/*                                                                     */
/***********************************************************************/

int main(int argc, char *argv[])
{
  par_type pars, *parp = &pars;
  dat_type dats, *datp = &dats;
  aux_type auxs, *auxp = &auxs;
  int      errflg;
  char     c;

  /* set defaults for options */
  parp->v_hyp      = -DBL_MAX;   /* unlimited hyperpolarization   */
  parp->v_res      = 0.0;        /* reset to resting potential    */
  parp->reset_t    = 0;          /* do not reset after each spike */
  parp->dt_min     = 1e-9;       /* minimum time step             */
  parp->max_spikes = 10000L;
  parp->do_trace   = 0;          /* no trace                      */
  parp->tracename  = NULL;
  parp->seed       = 0L;         /* use clock as seed             */

  /* process options */
  errflg = 0;
  opterr = 0;   /* errors handled by program */
  
  while ( !errflg && ( c = getopt(argc, argv, ":d:hy:r:RM:T:s:") ) != -1 )
    switch ( c ) {
    case 'd':      /* min. time step    */
      sscanf(optarg, "%lf", &parp->dt_min);
      break;
    case 'y':      /* hyperpolarization */
      sscanf(optarg, "%lf", &parp->v_hyp);
      break;
    case 'r':      /* reset potential   */
      sscanf(optarg, "%lf", &parp->v_res);
      break;
    case 'R':      /* time reset        */
      parp->reset_t = 1;
      break;
    case 'M':      /* max num. spikes   */
      sscanf(optarg, "%ld", &parp->max_spikes);
      break;
    case 'T':      /* trace             */
      parp->do_trace = 1;
      parp->tracename = optarg;
      break;
    case 's':      /* random seed       */
      sscanf(optarg, "%ld", &parp->seed);
      break;
    case 'h':      /* help              */
      errflg++;
      break;
    case ':':      /* missing option    */
      fprintf(stderr, "Option -%c requires an argument\n", optopt);
      errflg++;
      break;
    case '?':      /* wrong option      */
      fprintf(stderr, "Unrecognized option -%c\n", optopt);
      errflg++;
      break;
    default:
      fprintf(stderr, "This error is impossible\n");
      errflg++;
      break;
    }   /* switch, while */

  /* now read the required input arguments if okay */
  if ( optind + 5 > argc || errflg )
    {
      fprintf(stderr, USAGE);
      exit(-1);
    }

  sscanf(argv[optind++], "%lf", &parp->mean);
  sscanf(argv[optind++], "%lf", &parp->sigma);
  sscanf(argv[optind++], "%lf", &parp->rms);
  parp->inname  = argv[optind++];
  parp->outname = argv[optind++];

  init_calc(parp, datp, auxp);
  do_calc(parp, datp, auxp);
  save_calc(parp, datp);
  clear_calc(parp, datp, auxp);

  return 0;
}


/***********************************************************************/
/*                                                                     */
/*                  functions                                          */
/*                                                                     */
/***********************************************************************/

void init_calc(par_type *parp, dat_type *datp, aux_type *auxp)
{ 
  double   *tmp;
  MATFile  *infile;
  char     inname[FILE_NAME_LENGTH];

  /* read stimulus from file */
  sprintf(inname, "%s.mat", parp->inname);
  infile = matOpen(inname, "r");
  if ( !infile )
    {
      fprintf(stderr, "\n Error: Could not matOpen %s!\n\n", parp->inname);
      exit(-11);
    }
  
  /* get stimulus time step, integrated stimulus values */
  myMatGet(infile, "dt", &tmp);
  parp->n_in = myMatGet(infile, "F_n", &(parp->iS));

  matClose(infile);
  
  if ( !tmp || !parp->iS )  
    {
      fprintf(stderr, "\n Error: Could not read from %s!\n\n", parp->inname);
      exit(-12);
    }
  else
    {
      parp->dt_in = *tmp;
      free(tmp);
    }

  /* time limit of stimulus */
  auxp->t_max = (parp->n_in - 1) * parp->dt_in;

  /* allocate memory for spikes and check */
  datp->t_spike = (double *) calloc(parp->max_spikes, sizeof(double));
  if ( !datp->t_spike )
    {
      fprintf(stderr, "\n Could not allocate memory!\n\n");
      exit(-13);
    }
  
  /* prepare trace recording if wanted */
  if ( parp->do_trace )
    {
      parp->dt_trace = parp->dt_in;   /* same time step as stimulus */

      if ( parp->reset_t )
	auxp->n_max = parp->max_spikes*(long)ceil(auxp->t_max/parp->dt_trace);
      else
	auxp->n_max = (long) ceil( auxp->t_max / parp->dt_trace );
      /* two extra points for each spike */
      auxp->n_max += 2 * parp->max_spikes;

      /* allocate memory for data and auxiliary arrays */
      datp->t_trace = (double *) calloc(auxp->n_max, sizeof(double));
      datp->v_trace = (double *) calloc(auxp->n_max, sizeof(double));
      /* check allocation */
      if (    (!datp->t_trace && auxp->n_max > 0 )
           || (!datp->v_trace && auxp->n_max > 0 ) )
	{
	  fprintf(stderr, "\n Could not allocate memory!\n\n");
	  exit(-13);
	}
    }
  else
    auxp->n_max = 0L;

  /* variance and voltage limits corresponding to max and min time step */
  auxp->sig_min = parp->rms * sqrt(- 0.5 * expm1(-2.0*parp->dt_min));
  auxp->sig_max = parp->rms * sqrt(- 0.5 * expm1(-2.0*parp->dt_in ));
  auxp->v_min   = 1.0 - THRESHOLD_DISTANCE * auxp->sig_min;
  auxp->v_max   = 1.0 - THRESHOLD_DISTANCE * auxp->sig_max;
  auxp->v_min   = max(auxp->v_min, 0.0);
  auxp->v_max   = max(auxp->v_max, 0.0);

  /* initialize */
  datp->n_spike = 0L;
  datp->n_trace = 0L;

  /* seed random number generator */
  init_rand55(parp->seed);

  return;
}

/***********************************************************************/

void do_calc(par_type *p, dat_type *d, aux_type *a)

     /* This routine performs the actual computations. */

{
  double t_spike;

  /* initial spike at t = 0 */
  d->t_spike[d->n_spike++] = 0.0;

  /* compute spikes until number of spikes reached or time is up */
  while ( d->n_spike < p->max_spikes && (t_spike = next_spike(p, d, a)) )
    d->t_spike[d->n_spike++] = t_spike;

  return;

}  /* end -- do_calc */

/***********************************************************************/

void save_calc(par_type *p, dat_type *d)
{
  MATFile  *outfile;
  double   tmp;
  char     outname[FILE_NAME_LENGTH];

  sprintf(outname, "%s.mat", p->outname);
  outfile = matOpen(outname, "w");
  if ( !outfile )
    {
      fprintf(stderr, "\nCould not matOpen %s !\n", p->outname);
      exit(1);
    }

  tmp = (double) d->n_spike;
  myMatPut(outfile, "n_spike", 1, 1, &tmp);
  myMatPut(outfile, "t_spike", d->n_spike, 1, d->t_spike);

  matClose(outfile);

  if ( p->do_trace )
    {
      sprintf(outname, "%s.mat", p->tracename);
      outfile = matOpen(outname, "w");
      if ( !outfile )
	{
	  fprintf(stderr, "\nCould not matOpen %s !\n", p->outname);
	  exit(1);
	}

      tmp = (double) d->n_trace;
      myMatPut(outfile, "n_trace", 1, 1, &tmp);   
      myMatPut(outfile, "t_trace", d->n_trace, 1, d->t_trace);
      myMatPut(outfile, "v_trace", d->n_trace, 1, d->v_trace);

      matClose(outfile);
    }

  return;
}

/***********************************************************************/

void clear_calc(par_type *parp, dat_type *datp, aux_type *auxp)

{
  free(datp->t_spike);
  free(parp->iS);
  if ( parp->do_trace )
    {
      free(datp->t_trace);
      free(datp->v_trace);
    }
  return;
}

/***********************************************************************/

double next_spike(par_type *p, dat_type *d, aux_type *a)

/* this routine actually the membrane potential from
     reset potential to treshold crossing
   returns 
      time of spike    if threshold reached
      0                otherwise (time limit, trace mem full)         */

{
  double v,        /* potential, reset on entry */
         t,        /* time                      */
         x,        /* relative dist from thresh */
         sig,      /* actually used variance    */
         dt,       /* actual time step          */
         t_next,   /* time of next trace record */
         dx, tmp,  /* for interpolating in iS   */
         iS_old,   /* iS at t                   */
         iS_new;   /* iS at t+dt                */
  long   j;

  /* reset potential, set time to time of last spike or t=0 */
  v = p->v_res;
  if ( p->reset_t )
    {
      t = 0.0;
      iS_old = p->iS[0];
    }
  else
    {
      t = d->t_spike[d->n_spike-1];
      /* interpolate iS_old value */ 
      dx = modf(t/p->dt_in, &tmp); 
      j  = (long) tmp;
      iS_old = p->iS[j] + dx * (p->iS[j+1] - p->iS[j]);
    }

  /* starting point of trace */
  if ( p->do_trace )
    {
      d->t_trace[d->n_trace] = t;
      d->v_trace[d->n_trace] = v;
      t_next = t + p->dt_trace;
      if ( ++d->n_trace >= a->n_max )  /* stop */
	return 0.0;
    }
  else
    t_next = DBL_MAX;    /* never ever get here */

  /* iterate until threshold reached or tmax */
  while ( v < 1.0 && t < a->t_max )
    {
      /* get step length for next step, Eq. (B.4)
         note: a->v_max < a->v_min < 1 */
      if ( v < a->v_max )
	{                   /* far from thresh, max step size */
	   dt = p->dt_in;    
	  sig = a->sig_max;
	}
      else if ( v > a->v_min )
	{                   /* very close, min step size              */
	   dt = p->dt_min;  
	  sig = a->sig_min;
	}
      else 
	{                   /* close to thresh, adjust step size  */
            x = ( 1.0 - v ) / p->rms / THRESHOLD_DISTANCE;
	   dt = - 0.5 * log1p( -2.0 * x*x );
	  sig = (1.0-v) / THRESHOLD_DISTANCE;
	}

      /* perform step, Eq. (B.3)  */
      /* interpolate new iS value */ 
      dx = modf(t/p->dt_in, &tmp); j = (long) tmp;
      iS_new = p->iS[j] + dx * (p->iS[j+1] - p->iS[j]);

      t += dt;
      v  = v * exp(-dt) - p->mean * expm1(-dt) 
                        + p->sigma * (iS_new - iS_old * exp(-dt)) 
                        + sig * gauss_rand55();
      iS_old = iS_new;

      /* enforce hyperpolarization limit */
      v = max(v, p->v_hyp);

      /* trace if required, t_next==DBL_MAX if trace unwanted */
      if ( t > t_next )
	{
	  d->t_trace[d->n_trace] = t;
	  d->v_trace[d->n_trace] = v;
	  t_next = t + p->dt_trace;
	  if ( ++d->n_trace >= a->n_max )  /* stop */
	    return 0.0;
	}
      
    }

  /* check if we got a spike */
  if ( v >= 1.0 )
    {
      /* trace if required */
      if ( p->do_trace )
	{
	  d->t_trace[d->n_trace] = t;
	  d->v_trace[d->n_trace] = v;
	  if ( ++d->n_trace >= a->n_max )  /* stop */
	    return 0.0;
	}
      return t;
    }
  else
    return 0.0;

  /* if all else fails */
  return 0.0;

}

/***********************************************************************/

/*  Now follow MATLAB matrix handling functions                        */

/***********************************************************************/

long myMatGet(MATFile *fp, const char *name, double **data)

/* places r x c matrix under name 'name' in file fp, takes data from data */

{
  long    rows, cols;
  mxArray *mdata;

  /* read MATLAB data structure */
  mdata = matGetArray(fp, name);
  if ( !mdata )
    {
      *data = NULL;   /* read error */
      return 0;
    }

  rows = mxGetM(mdata);
  cols = mxGetN(mdata);
  if ( rows > 1 && cols > 1 )
    {
      *data = NULL;
      return 0;
    }

  if ( cols > 1 )
    rows = cols;

  /* allocate memory, works for 1D only!!! */
  *data = malloc(rows*sizeof(double));
  if ( !*data )
    return 0;

  /* copy data */
  memcpy(*data, mxGetPr(mdata), rows*sizeof(double));

  mxDestroyArray(mdata);

  return rows;
}  /* end -- myMatGet */


/**********************************************************************/

void myMatPut(MATFile *fp, char *name, long rows, long cols, double *data)

/* places r x c matrix under name 'name' in file fp, takes data from data */

{
  mxArray *mdata;
 
  mdata = mxCreateDoubleMatrix(rows, cols, mxREAL);
  memcpy(mxGetPr(mdata), data, rows*cols*sizeof(double));
  mxSetName(mdata, name);
  matPutArray(fp, mdata);
  mxDestroyArray(mdata);

  return;
}  /* end -- myMatPut */







      
      
