/*

  p_train.c   Simulate spike train in response to periodic stimulus
                Hyperpolarization can be limited
                Reset potential can be different from resting potential
		Time can be reset after each spike
		If desired, a file containing a voltage trace is saved
                All file I/O with MATLAB files

    	        Literature: Hans E. Plesser, Dissertation (1999), Appendix B
                            D. Gillespie, Phys Rev E 54:2084-2091 (1996)


  Copyright (C) 2000 Hans Ekkehard Plesser

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


  The calling syntax is:
*/


#define USAGE \
"                                                                      \n"\
" Simulation of spike trains for periodic stimuli                      \n"\
"                                                                      \n"\
"  p_train [options] mu q Om rms outfile                               \n"\
"                                                                      \n"\
"  Input:         mu  DC input                                         \n"\
"                  q  amplitude of AC input			       \n"\
"                 Om  stimulus frequency / pi	[ in radian ]	       \n"\
"                rms  rms amplitude of noise			       \n"\
"            outfile  MATLAB file containing n_spike t_spike	       \n"\
"								       \n"\
"  Options:   -d ###  minimum step size [ default 1e-9 ]               \n"\
"             -D ###  maximum step size [ min{0.01, 100 steps/period} ]\n"\
"             -y ###  minimum potential [ unbounded hyperpolarization ]\n"\
"             -r ###  reset potential   [ 0 ]    		       \n"\
"             -R      reset time after each spike  [ do not reset ]    \n"\
"             -A ###  absolute refractory time [ 0 ]                   \n"\
"             -p ###  initial phase / phase to reset to, degrees [ 0 ] \n"\
"             -t ###  membrane time constant in s [ 1 ]                \n"\
"             -i ###  maximum ISI length, skip      [ unbounded ]      \n"\
"             -I ###  maximum ISI length, terminate [ unbounded ]      \n"\
"             -M ###  max. number of spikes [ 10000 ]		       \n"\
"             -N ###  number stimulus repetitions [ 1 ]		       \n"\
"             -L ###  length of individual presentation [ Inf ]        \n"\
"             -T xxx  write potential trace to file xxx   [ no trace ] \n"\
"             -s ###  random generator seed [ use clock ]              \n"\
"             -Z      gzip output file [ do not gzip ]                 \n"\
"                                                                      \n"\
"  Notes:     Constrained stimulation:                                 \n"\
"             * -R forces time reset to t = 0 (abs refr time) after    \n"\
"               each spike.                                            \n"\
"             * If option -i ### is used, simulation of each interval  \n"\
"               is terminated if no spike occured after ### time units.\n"\
"               The number of intervals exceeding ### is then number   \n"\
"               of spikes 'missing' in the output file.                \n"\
"             * The output file contains number of spikes in n_spike   \n"\
"               and interspike-intervals in t_spike.                   \n"\
"                                                                      \n"\
"             Unconstrained stimulation:                               \n"\
"             * If option -I ### is used, the simulation will be       \n"\
"               terminated after more than ### time units have passed  \n"\
"               without a spike.                                       \n"\
"             * If option -N ### requires -L *** and ignores -I $$$.   \n"\
"               In this case, the stimulus is repeated ### times for a \n"\
"               time of ***.  Time startes anew at t=0 in each repeti- \n"\
"               tion, but is not reset after spikes.  This option may  \n"\
"               be useful to compile ISI histograms from a single call \n"\
"               of p_train.                                            \n"\
"             * Option -L ### is ignored if not used with -N ###.      \n"\
"             * The output file contains number of spikes in n_spike   \n"\
"               and absolute spike times in t_spike.                   \n"\
"                                                                      \n"\
"             Voltage trace:                                           \n"\
"             * Option -T is EXTREMELY memory-hungry, since the trace  \n"\
"               is buffered in memory during simulation. Use with care!\n"\
"             * Trace file contains number of trace points in n_trace, \n"\
"               and v_trace(t_trace) is voltage trace                  \n"\
"                                                                      \n"
       
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <string.h>
#include <getopt.h>

#include <rand55.h>   /* random number generator */

#include <mat.h>      /* MATLAB matrix handling  */

/***********************************************************************/
/*                                                                     */
/*                  Macros                                             */
/*                                                                     */
/***********************************************************************/

#define	max(A, B)	((A) > (B) ? (A) : (B))
#define	min(A, B)	((A) < (B) ? (A) : (B))
#define odd(n)          ( (n) & 1 )

#define pi     3.14159265358979324
#define root2  1.41421356237309505

#define THRESHOLD_DISTANCE 5.0   /* for choice of dt, see Eq. (B.4) */

#define FILE_NAME_LENGTH   255

/***********************************************************************/
/*                                                                     */
/*                  data structures                                    */
/*                                                                     */
/***********************************************************************/

typedef struct              /* input/problem parameters      */
{
  long   max_spikes,        /* maximum number of spikes      */
         no_repets,         /* number of stim. repetitions   */
         seed;              /* random generator seed         */
  int    reset_t,           /* reset to t=0 after each spike */
         do_repeat,         /* repeat stimulus               */
         do_terminate,      /* terminate if max ISI exceeded */
         do_trace,          /* record membrane pot. trace    */
         do_zip;            /* gzip output                   */
  double dt_min,            /* minimum time step             */
         dt_max,            /* maximum time step             */
         dt_trace,          /* minimum step for volt. tr.    */
         stim_length,       /* stim. duration for repetit.   */
         mu,                /* DC input                      */
         q,                 /* AC amplitude                  */
         om,                /* stimulus frequency            */
         phi,               /* initial phase                 */
         rms,               /* noise amplitude               */
         tau_m,             /* membrane time constant        */
         t_refrac,          /* absolute refractory time      */
         max_isi,           /* maximum ISI duration          */
         v_hyp,             /* minimum hyperpolarization     */
         v_res;             /* reset potential               */
  char   *inname,           /* input file name               */
         *outname,          /* name of output file           */
         *tracename;        /* name of voltage trace file    */
} par_type;

typedef struct              /* spike train, volt. trace      */
{
  long    n_spike,          /* number of spikes so far       */
          n_trace;          /* number of trace entries       */
  double *t_spike,          /* spike times                   */
         *t_trace,          /* time axis of voltage trace    */
         *v_trace;          /* voltage trace                 */
} dat_type;

typedef struct              /* auxiliary quantities          */
{
  long    n_max;            /* length of trace array         */
  double  eta,              /* effective phase               */
          q_eff,            /* effective AC amplitude        */
          v_max,            /* dt_in  for v < v_max, Eq.(B.4)*/
          v_min,            /* dt_min for v > v_min          */
          sig_min,          /* corresponding variances       */
          sig_max;
} aux_type;

/***********************************************************************/
/*                                                                     */
/*                           Functions                                 */
/*                                                                     */
/***********************************************************************/

void init_calc(par_type *parp, dat_type *datp, aux_type *auxp);
void do_calc(par_type *p, dat_type *d, aux_type *a);
void save_calc(par_type *p, dat_type *d);
void clear_calc(par_type *p, dat_type *d, aux_type *a);

double next_spike(par_type *p, dat_type *d, aux_type *a);

/* simplify MATLAB matrix handling */
long myMatGet(MATFile *fp, const char *name, double **data);
void myMatPut(MATFile *fp, char *name, long rows, long cols, double *data);


/***********************************************************************/
/*                                                                     */
/*                  main program                                       */
/*                                                                     */
/***********************************************************************/

int main(int argc, char *argv[])
{
  par_type pars, *parp = &pars;
  dat_type dats, *datp = &dats;
  aux_type auxs, *auxp = &auxs;
  int      errflg;
  char     c;

  /* set defaults for options */
  parp->phi        = 0.0;        /* start from phase 0            */
  parp->tau_m      = 1.0;        /* use scaled time units         */
  parp->t_refrac   = 0.0;        /* absolute refractory time      */
  parp->max_isi    = DBL_MAX;    /* unlimited ISI duration        */
  parp->do_repeat  = 0;          /* stimulus not repeated by def. */
  parp->no_repets  = 1;          /* single stimulus presentation  */
  parp->stim_length= DBL_MAX;    /* infinite stimulus             */
  parp->do_terminate= 1;         /* terminate if max_isi exceeded */
  parp->v_hyp      = -DBL_MAX;   /* unlimited hyperpolarization   */
  parp->v_res      = 0.0;        /* reset to resting potential    */
  parp->reset_t    = 0;          /* do not reset after each spike */
  parp->dt_min     = 1e-9;       /* minimum time step             */
  parp->dt_max     = -1.0;       /* dummy value, see init_calc    */
  parp->max_spikes = 10000L;
  parp->do_trace   = 0;          /* no trace                      */
  parp->tracename  = NULL;
  parp->seed       = 0L;         /* use clock as seed             */
  parp->do_zip     = 0;          /* do not gzip output            */

  /* process options */
  errflg = 0;
  opterr = 0;   /* errors handled by program */
  
  while (   !errflg 
         && ( c=getopt(argc,argv,":t:A:d:D:i:I:hy:r:Rp:M:N:L:T:s:Z") ) != -1 )
    switch ( c ) {
    case 't':      /* memb. time const. */
      sscanf(optarg, "%lf", &parp->tau_m);
      break;
    case 'A':      /* abs. refr. time   */
      sscanf(optarg, "%lf", &parp->t_refrac);
      break;
    case 'd':      /* min. time step    */
      sscanf(optarg, "%lf", &parp->dt_min);
      break;
    case 'D':      /* max. time step    */
      sscanf(optarg, "%lf", &parp->dt_max);
      break;
    case 'i':      /* max. ISI length, skip        */ 
      sscanf(optarg, "%lf", &parp->max_isi);
      parp->do_terminate = 0;
      break;
    case 'I':      /* max. ISI length, terminate   */
      sscanf(optarg, "%lf", &parp->max_isi);
      parp->do_terminate = 1;
      break;
    case 'y':      /* hyperpolarization limit      */
      sscanf(optarg, "%lf", &parp->v_hyp);
      break;
    case 'r':      /* reset potential   */
      sscanf(optarg, "%lf", &parp->v_res);
      break;
    case 'R':      /* time reset        */
      parp->reset_t = 1;
      break;
    case 'p':      /* initial phase     */
      sscanf(optarg, "%lf", &parp->phi);
      break;
    case 'M':      /* max num. spikes   */
      sscanf(optarg, "%ld", &parp->max_spikes);
      break;
    case 'N':      /* num. stim repets  */
      sscanf(optarg, "%ld", &parp->no_repets);
      parp->do_repeat = parp->no_repets > 0;
      break;
    case 'L':      /* stim length       */
      sscanf(optarg, "%lf", &parp->stim_length);
      break;
    case 'T':      /* trace             */
      parp->do_trace = 1;
      parp->tracename = optarg;
      break;
    case 's':      /* random seed       */
      sscanf(optarg, "%ld", &parp->seed);
      break;
    case 'Z':      /* gzip output       */
      parp->do_zip = 1;
      break;
    case 'h':      /* help              */
      errflg++;
      break;
    case ':':      /* missing option    */
      fprintf(stderr, "Option -%c requires an argument\n", optopt);
      errflg++;
      break;
    case '?':      /* wrong option      */
      fprintf(stderr, "Unrecognized option -%c\n", optopt);
      errflg++;
      break;
    default:
      fprintf(stderr, "This error is impossible\n");
      errflg++;
      break;
    }   /* switch, while */

  /* skipping over-long intervals makes sense only with reset */
  if ( !parp->do_terminate && !parp->reset_t )
    {
      fprintf(stderr, "Option -i ### requires option -R\n");
      exit(-1);
    }

  /* check proper use of options when repeating */
  if ( parp->do_repeat && parp->stim_length == DBL_MAX )
    {
      fprintf(stderr, "Option -N ### requires option -L ###\n");
      exit(-1);
    }
  else if ( !parp->do_repeat && parp->stim_length < DBL_MAX )
    {
      fprintf(stderr, 
          "Warning: Option -L ### has no effect unless used with -N ###\n");
    }


  /* now read the required input arguments if okay */
  if ( optind + 5 > argc || errflg )
    {
      fprintf(stderr, USAGE);
      exit(-1);
    }

  sscanf(argv[optind++], "%lf", &parp->mu);
  sscanf(argv[optind++], "%lf", &parp->q);
  sscanf(argv[optind++], "%lf", &parp->om);
  sscanf(argv[optind++], "%lf", &parp->rms);
  parp->outname = argv[optind++];

  init_calc(parp, datp, auxp);
  do_calc(parp, datp, auxp);
  save_calc(parp, datp);
  clear_calc(parp, datp, auxp);

  return 0;
}


/***********************************************************************/
/*                                                                     */
/*                  functions                                          */
/*                                                                     */
/***********************************************************************/

void init_calc(par_type *parp, dat_type *datp, aux_type *auxp)
{ 

  /* convert parameters, scale times */
  parp->om       *= pi * parp->tau_m;
  parp->phi      *= pi / 180.0;
  parp->rms      /= sqrt(parp->tau_m);   
  parp->t_refrac /= parp->tau_m;
  parp->dt_min   /= parp->tau_m;   
  parp->stim_length /= parp->tau_m;
  if ( parp->dt_max > 0.0 )              /* dt_max set by option -D    */
    parp->dt_max /= parp->tau_m;
  else                                   /* 0.01 or 100 steps / period */
    parp->dt_max  = min(0.01, 2.0*pi / parp->om / 100);
  parp->dt_max = max(parp->dt_max, parp->dt_min*2.0); /* just be safe  */
  if ( parp->max_isi < DBL_MAX )
    parp->max_isi /= parp->tau_m;

  /* some auxiliary quantities: cf. Eq. (2.43)  */
  auxp->eta   = parp->phi + atan(1.0 / parp->om);
  auxp->q_eff = parp->q / sqrt(1.0 + parp->om*parp->om);

  /* allocate memory for spikes and check */
  datp->t_spike = (double *) calloc(parp->max_spikes, sizeof(double));
  if ( !datp->t_spike )
    {
      fprintf(stderr, "\n Could not allocate memory!\n\n");
      exit(-13);
    }
  
  /* prepare trace recording if wanted */
  if ( parp->do_trace )
    {
      parp->dt_trace = parp->dt_max;

      /* avoid floating point overflow on number of trace points */
      if ( log(parp->max_spikes)+log(parp->max_isi) 
             > log(parp->dt_trace) + log(DBL_MAX) - 2 )
	{
	  fprintf(stderr, "\n Too many points in trace!\n\n");
	  exit(-13);
	}

      /* calculate number of trace points */
      auxp->n_max    = 
               parp->max_spikes * (long)ceil(parp->max_isi/parp->dt_trace);

      /* two extra points for each spike */
      auxp->n_max += 2 * parp->max_spikes;

      /* allocate memory for data and auxiliary arrays */
      datp->t_trace = (double *) calloc(auxp->n_max, sizeof(double));
      datp->v_trace = (double *) calloc(auxp->n_max, sizeof(double));
      /* check allocation */
      if (    (!datp->t_trace && auxp->n_max > 0 )
           || (!datp->v_trace && auxp->n_max > 0 ) )
	{
	  fprintf(stderr, "\n Could not allocate memory for trace!\n\n");
	  exit(-13);
	}
    }
  else
    auxp->n_max = 0L;

  /* variance and voltage limits corresponding to max and min time step */
  auxp->sig_min = parp->rms * sqrt(- 0.5 * expm1(-2.0*parp->dt_min));
  auxp->sig_max = parp->rms * sqrt(- 0.5 * expm1(-2.0*parp->dt_max));
  auxp->v_min   = 1.0 - THRESHOLD_DISTANCE * auxp->sig_min;
  auxp->v_max   = 1.0 - THRESHOLD_DISTANCE * auxp->sig_max;
  auxp->v_min   = max(auxp->v_min, 0.0);
  auxp->v_max   = max(auxp->v_max, 0.0);

  /* initialize */
  datp->n_spike = 0L;
  datp->n_trace = 0L;

  /* seed random number generator */
  init_rand55(parp->seed);

  return;
}

/***********************************************************************/

void do_calc(par_type *p, dat_type *d, aux_type *a)

     /* This routine performs the actual computations. */

{
  double t_spike;

  /* split between non-repeated and repeated stimuli */
  if ( !p->do_repeat )
    {
      /* initial spike at t = 0 */
      d->t_spike[d->n_spike++] = 0.0;
      
      /* compute spikes until number of spikes reached or time is up */
      /*   t_spike returns  0 to signal abort 
           t_spike returns -1 to signal interval > max_ISI without spike,
                                        to be ignored (option -i)    */
      while ( d->n_spike < p->max_spikes && (t_spike = next_spike(p, d, a)) )
	if ( t_spike > 0.0 )
	  d->t_spike[d->n_spike++] = t_spike * p->tau_m;  /* convert time */
	else
	  p->max_spikes--;
    }
  else
    while ( p->no_repets-- && d->n_spike < p->max_spikes )
      {
	/* initial spike at t = 0 */
	d->t_spike[d->n_spike++] = 0.0;
      
	/* compute spikes until number of spikes reached or time is up */
	/*   ignore spikes at negative times, but deduct them from
	     total number of spikes                                    */
	while (     d->n_spike < p->max_spikes 
                && (t_spike = next_spike(p, d, a)) > 0.0 )
          d->t_spike[d->n_spike++] = t_spike * p->tau_m;  /* convert time */
      }

  return;

}  /* end -- do_calc */

/***********************************************************************/

void save_calc(par_type *p, dat_type *d)
{
  MATFile  *outfile;
  double   tmp;
  char     outname[FILE_NAME_LENGTH],
           zip_command[FILE_NAME_LENGTH+20];

  sprintf(outname, "%s.mat", p->outname);
  outfile = matOpen(outname, "w");
  if ( !outfile )
    {
      fprintf(stderr, "\nCould not matOpen %s !\n", p->outname);
      exit(1);
    }

  tmp = (double) d->n_spike;
  myMatPut(outfile, "n_spike", 1, 1, &tmp);
  myMatPut(outfile, "t_spike", d->n_spike, 1, d->t_spike);

  matClose(outfile);

  if ( p->do_zip )
    {
      sprintf(zip_command, "gzip -f -q %s", outname);
      system(zip_command);
    }

  if ( p->do_trace )
    {
      sprintf(outname, "%s.mat", p->tracename);
      outfile = matOpen(outname, "w");
      if ( !outfile )
	{
	  fprintf(stderr, "\nCould not matOpen %s !\n", p->outname);
	  exit(1);
	}

      tmp = (double) d->n_trace;
      myMatPut(outfile, "n_trace", 1, 1, &tmp);   
      myMatPut(outfile, "t_trace", d->n_trace, 1, d->t_trace);
      myMatPut(outfile, "v_trace", d->n_trace, 1, d->v_trace);

      matClose(outfile);

      if ( p->do_zip )
	{
	  sprintf(zip_command, "gzip -f -q %s", outname);
	  system(zip_command);
	}

    }

  return;
}

/***********************************************************************/

void clear_calc(par_type *parp, dat_type *datp, aux_type *auxp)

{
  free(datp->t_spike);
  if ( parp->do_trace )
    {
      free(datp->t_trace);
      free(datp->v_trace);
    }
  return;
}

/***********************************************************************/

double next_spike(par_type *p, dat_type *d, aux_type *a)

/* this routine simulates the membrane potential from
     reset potential to treshold crossing
   returns 
      time of spike    if threshold reached
      -1               if max_ISI exceeded, but this should be ignored (-i)
      0                otherwise (time limit, trace mem full)         */

{
  double v,        /* potential, reset on entry */
         t,        /* time                      */
         t_max,    /* max time for this ISI     */
         x,        /* relative dist from thresh */
         sig,      /* actually used variance    */
         dt,       /* actual time step          */
         t_next,   /* time of next trace record */
         sin_old,  /* sin at t                  */
         sin_new;  /* sin at t+dt               */

  /* reset potential, set time to time of last spike or t=0 */
  v = p->v_res;
  if ( p->reset_t )
    t = p->t_refrac;
  else
    /* spike times are saved in real units, so scale */
    t = d->t_spike[d->n_spike-1] / p->tau_m + p->t_refrac;  
  sin_old = sin(p->om * t + a->eta);

  /* time limit for inter-spike-interval */
  if ( !p->do_repeat )
    t_max = p->max_isi < DBL_MAX ? t + p->max_isi : DBL_MAX;
  else
    t_max = p->stim_length;

  /* starting point of trace */
  if ( p->do_trace )
    {
      d->t_trace[d->n_trace] = t * p->tau_m;   /* save in real units */
      d->v_trace[d->n_trace] = v;
      t_next = t + p->dt_trace;
      if ( ++d->n_trace >= a->n_max )  /* stop */
	return 0.0;
    }
  else
    t_next = DBL_MAX;    /* never get to recording of trace point */

  /* iterate until threshold reached or max ISI duration reached */
  while ( v < 1.0 && t < t_max )
    {
      /* get step length for next step, Eq. (B.4) */
      if ( v < a->v_max )
	{                   /* far from thresh, max step size */
	   dt = p->dt_max;    
	  sig = a->sig_max;
	}
      else if ( v > a->v_min )
	{                   /* very close, min step size              */
	   dt = p->dt_min;  
	  sig = a->sig_min;
	}
      else 
	{                   /* close to thresh, adjust step size  */
            x = ( 1.0 - v ) / p->rms / THRESHOLD_DISTANCE;
	   dt = - 0.5 * log1p( -2.0 * x*x );
	  sig = (1.0-v) / THRESHOLD_DISTANCE;
	}

      /* perform step Eq. (B.3) */
      t += dt;
      sin_new = sin(p->om * t + a->eta);
      v  = v * exp(-dt) - p->mu * expm1(-dt) 
                        + a->q_eff * (sin_new - sin_old * exp(-dt)) 
                        + sig * gauss_rand55();
      sin_old = sin_new;

      /* enforce hyperpolarization limit */
      v = max(v, p->v_hyp);

      /* trace if required, t_next==DBL_MAX if trace unwanted */
      if ( t > t_next )
	{
	  d->t_trace[d->n_trace] = t * p->tau_m;   /* save in real units */
	  d->v_trace[d->n_trace] = v;
	  t_next = t + p->dt_trace;
	  if ( ++d->n_trace >= a->n_max )  /* stop */
	    return 0.0;
	}
      
    }

  /* check if we got a spike */
  if ( v >= 1.0 )
    {
      /* trace if required */
      if ( p->do_trace )
	{
	  d->t_trace[d->n_trace] = t * p->tau_m;  /* save in real units */
	  d->v_trace[d->n_trace] = v;
	  if ( ++d->n_trace >= a->n_max )  /* stop */
	    return 0.0;
	}
      return t;
    }
  else if ( !p->do_terminate && t >= t_max )
    return -1.0;   /* ignore ISI exceeding max_isi */
  else
    return 0.0;    /* stop                         */

  /* if all else fails */
  return 0.0;

}

/***********************************************************************/

/*  Now follow MATLAB matrix handling functions                        */

/***********************************************************************/

long myMatGet(MATFile *fp, const char *name, double **data)

/* places r x c matrix under name 'name' in file fp, takes data from data */

{
  long    rows, cols;
  mxArray *mdata;

  /* read MATLAB data structure */
  mdata = matGetArray(fp, name);
  if ( !mdata )
    {
      *data = NULL;   /* read error */
      return 0;
    }

  rows = mxGetM(mdata);
  cols = mxGetN(mdata);
  if ( rows > 1 && cols > 1 )
    {
      *data = NULL;
      return 0;
    }

  if ( cols > 1 )
    rows = cols;

  /* allocate memory, works for 1D only!!! */
  *data = malloc(rows*sizeof(double));
  if ( !*data )
    return 0;

  /* copy data */
  memcpy(*data, mxGetPr(mdata), rows*sizeof(double));

  mxDestroyArray(mdata);

  return rows;
}  /* end -- myMatGet */


/**********************************************************************/

void myMatPut(MATFile *fp, char *name, long rows, long cols, double *data)

/* places r x c matrix under name 'name' in file fp, takes data from data */

{
  mxArray *mdata;
 
  mdata = mxCreateDoubleMatrix(rows, cols, mxREAL);
  memcpy(mxGetPr(mdata), data, rows*cols*sizeof(double));
  mxSetName(mdata, name);
  matPutArray(fp, mdata);
  mxDestroyArray(mdata);

  return;
}  /* end -- myMatPut */







      
      
