function [chi, psi, sig, scale] = phase_dist(trains, Om, phi_0, nobins),

%    [chi, psi, sig, scale] = phase_dist(trains, Om, phi_0, nobins)
%
%    Firing Phase Distribution from spike train 
%
%    trains   columns are spike trains (absolute times, not ISIs)
%    Om       Om * pi is stimulus frequency [ radian ]
%    phi_0    stimulus phase at t=0         [ degree ]
%    nobins   number of bins
%
%    chi      distribution, normalized as discrete distribution
%    psi      phase axis, bin centers, radian
%    sig      std deviation of bin content
%    scale    scale * chi is spike count per bin

%  Copyright (C) 2000 Hans Ekkehard Plesser
%
%  This program is free software; you can redistribute it and/or
%  modify it under the terms of the GNU General Public License
%  as published by the Free Software Foundation; either version 2
%  of the License, or (at your option) any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program; if not, write to the Free Software
%  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

% get input size
[nospikes, notrains] = size(trains);

% range and binning of histogram, psi contains bin centers
psi_min = -pi;
dpsi    = 2*pi / nobins;
psi     = psi_min + 0.5 * dpsi + (0:(nobins-1)).' * dpsi;

% convert times to phases, shift to (-pi,pi]
phi_rad = phi_0 / 180 * pi;
phi = rem(Om * trains+phi_0, 2*pi);
ix  = find(phi > pi);
phi(ix) = phi(ix) - 2*pi;

% get bin indeces
bins = 1 + floor((phi - psi_min) / dpsi);

% build histogram
chi = zeros(nobins, notrains);
for bin = 1:nobins,
  chi(bin, :) = sum(bins == bin);
end

% number of valid spikes---should be all spikes here
N = repmat(sum(~isnan(bins)), [nobins, 1]);

% normalized histogram
chi   = chi ./ N;

if nargout > 2,
  % std dev of bin count, see Scott (1992)
  sig = sqrt(chi .* (1 - chi) ./ N);
end

if nargout > 3,
  % histogram scaling factor
  scale = N(1,:);
end

return



