function [S, omS, errS, nspike] = psd_train(train, n, k, Om, To, maxsec),

%   [S, omS, errS, nspike] = psd_train(train, n, k, Om, To [, maxsec])
%
%   computes power spectral density of spike train using FFT
%   for an observation time of To, i.e. spike train is split
%   in sections of length To
%
%   train   spike train (ordered vector of spike times)
%   n       FFT length is 2^n
%   k, Om   frequency Om*pi is put in bin k+1
%   To      compute spectrum for sections of length To
%   maxsec  maximum number of sections to use 
%
%   S       power spectral density
%   omS     frequency axis
%   errS    std error of mean of spectrum
%   nspike  number of spikes in each FFT section
%
%   discretization: dom = Om*pi/k, dt = 2/Om * k/2^n, To<=2/Om*k

%  Copyright (C) 2000 Hans Ekkehard Plesser
%
%  This program is free software; you can redistribute it and/or
%  modify it under the terms of the GNU General Public License
%  as published by the Free Software Foundation; either version 2
%  of the License, or (at your option) any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program; if not, write to the Free Software
%  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

% multiply Om by pi
Om = Om * pi;

fftlen = 2^n;
nyx    = fftlen/2+1;         % index of Nyquist frequency
dt     = 2*pi/Om * k/fftlen; % width of time bins
Tfft   = dt * fftlen;        % time length for FFT input
dom    = 2*pi / Tfft;        % frequency increment
omS    = (0:nyx-1).' * dom;  % frequency axis

if Tfft < To,
  error('FFT length too short.');
end

% make sure train starts at 0, discard spike at 0
train=train(2:end)-train(1);

Tmax = train(end);
nsec = floor(Tmax/To);       % this many sections can be formed 
if nargin > 5,               % use limit if given
  nsec = min([nsec maxsec]); 
end

if nsec == 0,
  S = repmat(NaN, nyx, 1);
  omS = S;
  errS = S;
  nspike = [];
  return
end

S = zeros(nyx, 1);     % accumulate spectrum over sections
S2= S;                 % accumulate square for std error

for j = 1:nsec,

  % find spikes that fit into section
  ix = find(0 < train(:) & train(:) <= To);

  % record number of spikes
  nspike(j) = length(ix);
  
  % generate spike vector for FFT
  % get time index
  tix = 1 + floor(floor(train(ix)/dt));
  
  % mark spikes with ones
  st = zeros(fftlen, 1);
  st(tix) = ones(size(tix));
  
  % shift times left by To to be ready for next section
  train = train - To;
  
  % compute FFT and convert to PSD
  % NOTE: the Ansatz is already for the one-sided PSD, so no
  %       factor of 2 here
  fst = fft(st);
  psd = fst(1:nyx) .* conj(fst(1:nyx));   
  
  % accumulate spectrum
  S  = S  + psd;
  S2 = S2 + psd.^2;

end

% average, std error of mean
S = S / nsec;
if nsec > 1,
  errS = sqrt((S2/nsec - S.^2)/(nsec-1));
else
  errS = repmat(NaN , size(S));
end

% normalize by observation time To
S    =    S / (pi*To);
errS = errS / (pi*To);

return
