/**************************************************************************
* Copyright 1995 Thomas Fricke
*
* Permission to use, copy, modify, distribute, and sell
* this software and its documentation for any purpose is
* hereby granted without fee, provided that the above copyright
* notice appear in all copies and that both that copyright notice
* and this permission notice appear in supporting documentation,
* and that the name of Thomas Fricke not be used in advertising or
* publicity pertaining to distribution of the software without
* specific, written prior permission.  Thomas Fricke makes no representations
* about the suitability of this software for any purpose.
* It is provided "as is" without express or implied warranty.
*
* THOMAS FRICKE DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
* INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
* IN NO EVENT SHALL THOMAS FRICKE BE LIABLE FOR ANY SPECIAL, INDIRECT OR
* CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
* LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
* NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
* CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*
* Author:
*        Thomas Fricke
*        Inst. f. Physik
*        Lehrstuhl fuer Stochastische Prozess
*        Humboldt Universitaet zu Berlin
*        Invalidenstr. 110
*        D - 10115 Berlin
*        GERMANY
*
*
*
*    A number of macros and procedures generating random numbers 
*    according the most important distributions.
*    
*    Based on:
*        Donald E. Knuth
*        The Art of Computer-Programming, Vol. II
*        Seminumerical Algorithms
*        Addison-Wesley, Massachusetts
*
*    The initialization:
*
*    init_rand55(seed):
*                    procedure randomizing the generator by a seed.
*                    The seed determines the pseudo random number
*                    sequence totally, same seed results in same sequence.
*        !!!         If not called at first, before using any function or
*                    macro, the sequence of random numbers is always the same !
*                    If seed is 0L, system time or clock function is used
*                    to imitate real randomness. The function returns the
*                    seed to allow to reproduce the sequence.
*
*    The uniform distributions:
*
*    drand55()    :    macro generating a uniform distributed double random 
*                    number in the range [0..1), 32 or (if the compiler
*                    does not convert ulong-->float correctly, like
*                    Atari TC 2.03) 31 bits are relevant.
*    lrand55(l)    :    macro generating an uniform distributed unsigned long
*                    random number in the range 0..l-1. Avoid use because
*                    the %-division wastes time. If possible use
*                    (rand55() & mask) instead!
*    rand55()    :    macro generating an uniform distributed unsigned long
*                    random number in the range 0..MAX_ULONG. The direct
*                    way!
*
*    The Alias-Method by A. J. Walker
*
*     init_alias55(double *prob, double *aliasprob, 
*                    int *aliaselse, const int nalias):
*
*                    input are three arrays of size nalias, the double array
*                    prob contains the (normalized) probabilities 
*                    of the random numbers to generate. The i'th entry of
*                    aliasprob is changed to the probability of the
*                    alias-method to accept i directly.
*                    The aliaselse array is set up to the alias to generate
*                    if i is not accepted. The possible range of i 
*        !!!            is 0..nalias-1,    where nalias MUST BE A POWER OF 2,
*                    else you have to fill up your array by 0s!
*                    
*    alias_rand55(aliasprob, aliaselse, nalias) :
*                    macro, whose inputs are the arrays aliasprob and aliaselse
*                    set up by init_alias55(prob,aliasprob,aliaselse,nalias).
*                    It generates long integer random numbers according the
*                    probability distribution prob you have called with 
*                    the procedure init_alias55().
*
*    Some important distributions: 
*
*    exp_rand55():
*                    macro, generating exponentially distributed double
*                    random numbers in the range [0..infinity) with mean
*                    1.0.
*
*    gauss_rand55():    function, using a modified Marsaglia-algorithm to
*                    generate gaussian distributed random numbers with
*                    mean 0 and standard deviation 1.
*/

#ifndef _RANDOM_H
#define _RANDOM_H 1

#ifdef __alpha 
#define _modpwr 64
#else
#define _modpwr 32 /*   Algorithm computes modulo 2**_modpwr  */
#endif
#define _K 55
#define _J 24

#define rand55() ( (_k55)?(_k55)--:((_k55)=54),\
                   (_j55)?(_j55)--:((_j55)=54),\
                   _s55[(_k55)]=_s55[(_k55)]+_s55[(_j55)])

#ifdef NONDP
#error "randmin not implemented"
#define exp_rand55 rand_min
#else
#define exp_rand55() ( - log( 1.0 - drand55() ) )
#endif

#ifdef __TURBOC__
#define drand55() ldexp((double)(rand55()>>1),1-_modpwr)
#else
#define drand55() ldexp((double)(rand55()),-_modpwr)
#endif

#define lrand55(l) ((rand55()>>1)%(l))

unsigned long init_rand55(unsigned long);

extern unsigned long int _sel55;
extern short     _j55,_0j55;
extern short     _k55,_0k55; 
/* --> Knuth, Art of Computer-Programming, Vol. 2, p. 172 */
extern unsigned long _s55[_K],_0s55[_K];

extern long     _alias55;
#define alias_rand55(aliasprob, aliaselse, nalias)\
            ( aliasprob[  _alias55 = ( rand55()&(nalias-1) )  ] > drand55()  ?\
                _alias55 : aliaselse[_alias55] )

void     init_alias55(double *prob, double *aliasprob,
                        int *aliaselse, const int nalias);

double gauss_rand55(void);

#endif
